# -*- coding: utf-8 -*-
# @Time    : 2023/10/14 下午3:10
# @Author  : sudoskys
# @File    : task.py
# @Software: PyCharm
from dotenv import load_dotenv
from llmkira.utils import sync
from loguru import logger
from pydantic import Field, BaseSettings, root_validator


class RabbitMQ(BaseSettings):
    """
    代理设置
    """
    amqp_dsn: str = Field("amqp://admin:admin@localhost:5672", env='AMQP_DSN')

    class Config:
        env_file = '.env'
        env_file_encoding = 'utf-8'

    @root_validator()
    def is_connect(cls, values):
        import aio_pika
        try:
            sync(aio_pika.connect_robust(
                values['amqp_dsn']
            ))
        except Exception as e:
            logger.error('RabbitMQ connect failed, pls set AMQP_DSN in .env')
            raise ValueError('RabbitMQ connect failed')
        else:
            logger.success(f"RabbitMQ connect success")
        return values


load_dotenv()
RabbitMQSetting = RabbitMQ()



# -*- coding: utf-8 -*-
# @Time    : 2023/8/18 下午7:08
# @Author  : sudoskys
# @File    : function.py
# @Software: PyCharm

__receiver__ = "llm_task"

import copy
import json
import os
import time

import shortuuid
from aio_pika.abc import AbstractIncomingMessage
from loguru import logger

from llmkira.middleware.chain_box import Chain, AUTH_MANAGER, CHAIN_MANAGER
from llmkira.schema import RawMessage
from llmkira.sdk.endpoint import openai
from llmkira.sdk.func_calling.register import ToolRegister
from llmkira.task import Task, TaskHeader


class ChainFunc(object):
    @staticmethod
    async def auth_chain(task: TaskHeader, func_name: str = "Unknown"):
        _task_forward: TaskHeader = task.copy()
        meta = _task_forward.task_meta.child(__receiver__)
        meta.continue_step += 1
        meta.callback_forward = False
        meta.callback_forward_reprocess = False
        meta.verify_uuid = shortuuid.uuid()[0:8]
        # 追加中断
        if meta.limit_child <= 0:
            return None
        _task_forward.task_meta = meta
        # 注册部署点
        task_id = await AUTH_MANAGER.add_auth(
            task=Chain(
                uuid=meta.verify_uuid,
                user_id=str(_task_forward.receiver.user_id),
                address=__receiver__,  # 重要：转发回来这里
                time=int(time.time()),
                arg=TaskHeader(
                    sender=_task_forward.sender,
                    receiver=_task_forward.receiver,
                    task_meta=meta,
                    message=[]
                )
            )
        )
        # 追加任务
        task_meta = copy.deepcopy(meta)
        task_meta.direct_reply = True
        await Task(queue=_task_forward.receiver.platform).send_task(
            task=TaskHeader(
                sender=task.sender,  # 继承发送者
                receiver=task.receiver,  # 因为可能有转发，所以可以单配
                task_meta=task_meta,
                message=[
                    RawMessage(
                        user_id=_task_forward.receiver.user_id,
                        chat_id=_task_forward.receiver.chat_id,
                        text=f"🔑 Type `/auth {task_id}` to confirm execution of function `{func_name}`"
                    )
                ]
            )
        )
        del task_meta
        return

    @staticmethod
    async def resign_chain(
            task: TaskHeader,
            parent_func: str,
            repeatable: bool
    ):
        """
        注册链
        :param task: 任务
        :param parent_func: 父函数
        :param repeatable: 是否可重复
        """
        _task_forward: TaskHeader = task.copy()
        meta = _task_forward.task_meta.child(__receiver__)
        meta.continue_step += 1
        meta.callback_forward = False
        meta.callback_forward_reprocess = False
        # 追加中断
        if meta.limit_child <= 0:
            return None
        _task_forward.task_meta = meta

        # 禁用子链使用出现过的函数
        # Repeatable
        try:
            if not repeatable:
                _task_forward.task_meta.function_list = [
                    item
                    for item in _task_forward.task_meta.function_list
                    if item.name != parent_func
                ]
        except Exception as e:
            logger.error(e)
            logger.warning(f"[362211]Remove function {parent_func} failed")
        # 注册部署点
        await CHAIN_MANAGER.add_task(
            task=Chain(
                user_id=str(_task_forward.receiver.user_id),
                address=_task_forward.receiver.platform,
                time=int(time.time()),
                arg=TaskHeader(
                    sender=_task_forward.sender,
                    receiver=_task_forward.receiver,
                    task_meta=meta,
                    message=[]
                )
            )
        )


class FunctionReceiver(object):
    """
    receive message from telegram
    """

    def __init__(self):
        self.task = Task(queue=__receiver__)

    async def deal_message(self, message: AbstractIncomingMessage):
        """
        处理message
        :param message:
        :return:
        """
        if os.getenv("LLMBOT_STOP_REPLY") == "1":
            return None
        # 解析数据
        _task: TaskHeader = TaskHeader.parse_raw(message.body)
        # 没有任何参数
        if not _task.task_meta.parent_call:
            return None

        _function: openai.OpenaiResult = openai.OpenaiResult.parse_obj(_task.task_meta.parent_call)
        func_message = _function.default_message
        if not func_message.function_call:
            return None
        logger.debug(f"[x] Received Function {func_message.function_call.name}")
        # 运行函数
        _arg = json.loads(func_message.function_call.arguments)
        _tool_cls = ToolRegister().get_tool(func_message.function_call.name)
        if not _tool_cls:
            logger.warning(f"Not found function {func_message.function_call.name}")
            return None
        _tool_obj = _tool_cls()
        if _tool_obj.require_auth:
            if not _task.task_meta.verify_uuid:
                await ChainFunc.auth_chain(task=_task, func_name=func_message.function_call.name)
                logger.warning(
                    f"[x] Function \n--auth-require {func_message.function_call.name} require."
                )
                return None
            else:
                _task.task_meta.verify_uuid = None

        # 追加步骤
        await ChainFunc.resign_chain(
            task=_task,
            parent_func=func_message.function_call.name,
            repeatable=_tool_obj.repeatable
        )
        # 运行函数
        await _tool_obj.run(
            task=_task,
            receiver=_task.receiver,
            arg=_arg
        )
        # 注册区域，必须在run之后
        """
        reload_tool = .get_tool(name=message.function_call.name)
        if reload_tool:
            logger.info(f"[875521]Chain child callback sent {message.function_call.name}")
            await reload_tool().callback(sign="resign", task=_task)
        """

    async def on_message(self, message: AbstractIncomingMessage):
        """
        处理message
        :param message:
        :return:
        """
        try:
            await self.deal_message(message=message)
        except Exception as e:
            logger.error(f"Function Receiver Error {e}")
            await message.reject(requeue=False)
        finally:
            await message.ack(multiple=False)

    async def function(self):
        logger.success("Receiver Runtime:Function Fork Cpu start")
        await self.task.consuming_task(self.on_message)



# -*- coding: utf-8 -*-


import cjieba
import numpy as np
from nltk.cluster.util import cosine_distance

from ..solo import singleton
from ..summarization import STOPWORDS as STOPWORDS_PATH

MIN_SEQ_LEN = 0


def load_stopwords(file_path):
    with open(file_path, encoding='utf-8') as f:
        return [line.strip() for line in f]


def split_doc(doc, stopwords=None):
    if not stopwords:
        stopwords = []

    sentences = []
    cut_sentences = []
    origin_sentences = []

    while len(doc) > 0:
        for i in range(len(doc)):
            if doc[i] in ['。', '！', '?', '？']:
                sentences.append(doc[:i + 1])
                doc = doc[i + 1:]
                break
    for sent in sentences:
        if len(sent) > MIN_SEQ_LEN:
            cut_sentences.append([word for word in cjieba.cut(sent) if word not in stopwords])
            origin_sentences.append(sent)
    return origin_sentences, cut_sentences


def sentence_similarity(sent1, sent2):
    """
    计算两个句子之间的相似性
    :param sent1:
    :param sent2:
    :return:
    """
    all_words = list(set(sent1 + sent2))

    vector1 = [0] * len(all_words)
    vector2 = [0] * len(all_words)

    for word in sent1:
        vector1[all_words.index(word)] += 1

    for word in sent2:
        vector2[all_words.index(word)] += 1

    # cosine_distance 越大越不相似
    return 1 - cosine_distance(vector1, vector2)


def build_similarity_matrix(sentences):
    """
    构建相似矩阵
    :param sentences:
    :return:
    """
    S = np.zeros((len(sentences), len(sentences)))
    for idx1 in range(len(sentences)):
        for idx2 in range(len(sentences)):
            if idx1 == idx2:
                continue
            S[idx1][idx2] = sentence_similarity(sentences[idx1], sentences[idx2])
    # 将矩阵正则化
    for idx in range(len(S)):
        if S[idx].sum == 0:
            continue
        S[idx] /= S[idx].sum()

    return S


def pagerank(A, eps=0.0001, d=0.85):
    P = np.ones(len(A)) / len(A)
    while True:
        new_P = np.ones(len(A)) * (1 - d) / len(A) + d * A.T.dot(P)
        delta = abs(new_P - P).sum()
        if delta <= eps:
            return new_P
        P = new_P


@singleton
class TextRankSummarization:
    def __init__(self, ratio):
        self.ratio = ratio
        self.stopwords = load_stopwords(STOPWORDS_PATH)

    def analysis(self, doc):
        origin_sentences, cut_sentences = split_doc(doc, stopwords=self.stopwords)

        S = build_similarity_matrix(cut_sentences)

        sentences_ranks = pagerank(S)

        sentences_ranks = [item[0] for item in sorted(enumerate(sentences_ranks), key=lambda item: -item[1])]

        selected_sentences_index = sorted(sentences_ranks[:int(len(origin_sentences) * self.ratio)])

        summary = []
        for idx in selected_sentences_index:
            summary.append(origin_sentences[idx])

        return ''.join(summary)



# coding:utf-8


class EditSimilarity:
    def __init__(self):
        pass

    def edit_dist(self, str1, str2):
        matrix = [[i + j for j in range(len(str2) + 1)] for i in range(len(str1) + 1)]

        for i in range(1, len(str1) + 1):
            for j in range(1, len(str2) + 1):
                if str1[i - 1] == str2[j - 1]:
                    d = 0
                else:
                    d = 1
                matrix[i][j] = min(matrix[i - 1][j] + 1, matrix[i][j - 1] + 1, matrix[i - 1][j - 1] + d)

        return matrix[len(str1)][len(str2)]



# -*- coding: utf-8 -*-
# @Time    : 2023/7/10 下午9:31
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
import hashlib
import pickle
from io import BytesIO
from typing import Union, List, Optional

import nest_asyncio
from pydantic import Field, BaseModel, validator
from telebot import types

from .sdk.schema import File

nest_asyncio.apply()


def generate_md5_short_id(data):
    # 检查输入数据是否是一个文件
    is_file = False
    if isinstance(data, str):
        is_file = True
    if isinstance(data, BytesIO):
        data = data.getvalue()
    # 计算 MD5 哈希值
    md5_hash = hashlib.md5()
    if is_file:
        with open(data, "rb") as file:
            for chunk in iter(lambda: file.read(4096), b""):
                md5_hash.update(chunk)
    else:
        md5_hash.update(data)
    # 获取哈希值的 16 进制表示
    hex_digest = md5_hash.hexdigest()
    # 生成唯一的短ID
    short_id = hex_digest[:8]
    return short_id


class RawMessage(BaseModel):
    user_id: int = Field(None, description="用户ID")
    chat_id: int = Field(None, description="群组ID")
    text: str = Field(None, description="文本")
    file: List[File] = Field([], description="文件")
    created_at: int = Field(None, description="创建时间")

    class Config:
        arbitrary_types_allowed = True
        extra = "allow"

    @validator("text")
    def check_text(cls, v):
        if v == "":
            v = "message is empty"
        if len(v) > 4096:
            v = v[:4090]
        return v

    @staticmethod
    async def download_file(file_id) -> Optional[File.Data]:
        from llmkira.cache.redis import cache
        file = await cache.read_data(file_id)
        if not file:
            return None
        file_obj: File.Data = pickle.loads(file)
        return file_obj

    @staticmethod
    async def upload_file(name, data):
        from llmkira.cache.redis import cache
        _key = str(generate_md5_short_id(data))
        await cache.set_data(key=_key, value=pickle.dumps(File.Data(file_name=name, file_data=data)),
                             timeout=60 * 60 * 24 * 7)
        return File(file_id=_key, file_name=name)

    @classmethod
    def from_telegram(cls, message: Union[types.Message, types.CallbackQuery]):
        if isinstance(message, types.Message):
            user_id = message.from_user.id
            chat_id = message.chat.id
            text = message.text
            created_at = message.date
        elif isinstance(message, types.CallbackQuery):
            user_id = message.from_user.id
            chat_id = message.message.chat.id
            text = message.data
            created_at = message.message.date
        else:
            raise ValueError(f"Unknown message type {type(message)}")
        return cls(
            user_id=user_id,
            text=text,
            chat_id=chat_id,
            created_at=created_at
        )


def singleton(cls):
    _instance = {}

    def inner():
        if cls not in _instance:
            _instance[cls] = cls()
        return _instance[cls]

    return inner



# -*- coding: utf-8 -*-
import logging
import os
from typing import Dict, Union

import fasttext
import requests

logger = logging.getLogger(__name__)
models = {"low_mem": None, "high_mem": None}
FTLANG_CACHE = os.getenv("FTLANG_CACHE", "/tmp/fasttext-langdetect")

try:
    # silences warnings as the package does not properly use the python 'warnings' package
    # see https://github.com/facebookresearch/fastText/issues/1056
    fasttext.FastText.eprint = lambda *args, **kwargs: None
except Exception as e:
    pass


def check_model(name: str) -> bool:
    # 查看模型是否为 0 字节
    target_path = os.path.join(FTLANG_CACHE, name)
    if os.path.exists(target_path):
        if os.path.getsize(target_path) > 0:
            return True
        else:
            # 0 字节，删除
            os.remove(target_path)
    return False


def download_model(name: str) -> str:
    target_path = os.path.join(FTLANG_CACHE, name)
    if not os.path.exists(target_path):
        logger.info(f"Downloading {name} model ...")
        url = f"https://dl.fbaipublicfiles.com/fasttext/supervised-models/{name}"  # noqa
        os.makedirs(FTLANG_CACHE, exist_ok=True)
        with open(target_path, "wb") as fp:
            response = requests.get(url)
            fp.write(response.content)
        logger.info(f"Downloaded.")
    return target_path


def get_or_load_model(low_memory=False):
    if low_memory:
        model = models.get("low_mem", None)
        if not model:
            check_model("lid.176.ftz")
            model_path = download_model("lid.176.ftz")
            model = fasttext.load_model(model_path)
            models["low_mem"] = model
        return model
    else:
        model = models.get("high_mem", None)
        if not model:
            check_model("lid.176.ftz")
            model_path = download_model("lid.176.bin")
            model = fasttext.load_model(model_path)
            models["high_mem"] = model
        return model


def detect(text: str, low_memory=False) -> Dict[str, Union[str, float]]:
    model = get_or_load_model(low_memory)
    labels, scores = model.predict(text)
    label = labels[0].replace("__label__", '')
    score = min(float(scores[0]), 1.0)
    return {
        "lang": label,
        "score": score,
    }



# -*- coding: utf-8 -*-
# @Time    : 2023/8/15 下午10:34
# @Author  : sudoskys
# @File    : base.py
# @Software: PyCharm
__version__ = "0.0.1"

from typing import Union, List, Optional, Literal

import httpx
from dotenv import load_dotenv
from llmkira.sdk.schema import Message, Function
from llmkira.utils import sha1_encrypt
from loguru import logger
from pydantic import BaseModel, root_validator, validator, Field, HttpUrl, BaseSettings

from .action import Tokenizer, TokenizerObj
from ...error import ValidationError
from ...network import request

load_dotenv()

MODEL = Literal[
    "gpt-3.5-turbo-0301",
    "gpt-3.5-turbo-0613",
    "gpt-3.5-turbo",
    "gpt-4-0314",
    "gpt-4-0613",
    "gpt-4"
]


class OpenaiResult(BaseModel):
    class Usage(BaseModel):
        prompt_tokens: int
        completion_tokens: int
        total_tokens: int

    class Choices(BaseModel):
        index: int
        message: Message = None
        finish_reason: str
        delta: dict = None

    id: str
    object: str
    created: int
    model: str
    choices: List[Choices]
    usage: Usage

    class Config:
        arbitrary_types_allowed = True
        extra = "allow"

    @property
    def result_type(self):
        return self.object

    @property
    def default_message(self):
        return self.choices[0].message


class Openai(BaseModel):
    class Proxy(BaseSettings):
        proxy_address: str = Field(None, env="OPENAI_API_PROXY")  # "all://127.0.0.1:7890"

        class Config:
            env_file = '.env'
            env_file_encoding = 'utf-8'
            case_sensitive = True
            arbitrary_types_allowed = True

    class Driver(BaseSettings):
        endpoint: HttpUrl = Field("https://api.openai.com/v1/chat/completions", env='OPENAI_API_ENDPOINT')
        api_key: str = Field(None, env='OPENAI_API_KEY')
        org_id: Optional[str] = Field(None, env='OPENAI_API_ORG_ID')

        # TODO:AZURE API VERSION

        # token: Tokenizer = TokenizerObj
        @validator("api_key")
        def check_key(cls, v):
            if v:
                if not str(v).startswith("sk-"):
                    logger.warning("OpenaiDriver:api_key should start with `sk-`")
                # if not len(str(v)) == 51:
                #    raise ValidationError("api_key must be 51 characters long")
            else:
                raise ValidationError("api_key is required,pls set OPENAI_API_KEY in .env")
            return v

        @property
        def uuid(self):
            # 取 api key 最后两位
            _flag = self.api_key[-2:]
            return f"{_flag}:{sha1_encrypt(self.api_key)}"

        class Config:
            env_file = '.env'
            env_file_encoding = 'utf-8'
            case_sensitive = True
            arbitrary_types_allowed = True
            extra = "allow"

    config: Driver
    model: MODEL = Field("gpt-3.5-turbo-0613", env='OPENAI_API_MODEL')
    messages: List[Message]
    functions: Optional[List[Function]] = None

    function_call: Optional[str] = None
    """
    # If you want to force the model to call a specific function you can do so by setting function_call: {"name": "<insert-function-name>"}. 
    # You can also force the model to generate a user-facing messages_box by setting function_call: "none". 
    # Note that the default behavior (function_call: "auto") is for the model to decide on its own whether to call a function and if so which function to call.
    """

    temperature: Optional[float] = 1
    top_p: Optional[float] = None
    n: Optional[int] = 1
    # Bot于流式响应负载能力有限
    stream: Optional[bool] = False
    stop: Optional[Union[str, List[str]]] = None
    max_tokens: Optional[int] = None
    presence_penalty: Optional[float] = None
    frequency_penalty: Optional[float] = None
    logit_bias: Optional[dict] = None
    user: Optional[str] = None

    # 用于调试
    echo: Optional[bool] = False

    def get_proxy_settings(self):
        proxy = self.Proxy()
        return proxy

    def update_model(self, model: MODEL = "gpt-3.5-turbo"):
        self.model = model

    @staticmethod
    def get_model_list():
        return MODEL.__args__

    @staticmethod
    def get_token_limit(model: str):
        v = 2048
        if "gpt-3.5" in model:
            v = 4096
        elif "gpt-4" in model:
            v = 8192
        if "-16k" in model:
            v = v * 4
        elif "-32k" in model:
            v = v * 4
        return v

    @validator("presence_penalty")
    def check_presence_penalty(cls, v):
        if not (2 > v > -2):
            raise ValidationError("presence_penalty must be between -2 and 2")
        return v

    @validator("stop")
    def check_stop(cls, v):
        if isinstance(v, list) and len(v) > 4:
            raise ValidationError("stop list length must be less than 4")
        return v

    @validator("temperature")
    def check_temperature(cls, v):
        if not (2 > v > 0):
            raise ValidationError("temperature must be between 0 and 2")
        return v

    @root_validator
    def check_root(cls, values):
        if values.get("functions"):
            if values.get("model") not in MODEL.__args__:
                raise ValidationError("function only support model: {}".format(MODEL.__args__))
        return values

    @staticmethod
    def parse_single_reply(response: dict) -> Message:
        """
        解析响应
        :param response:
        :return:
        """
        # check
        if not response.get("choices"):
            raise ValidationError("Message is empty")

        _message = Message.parse_obj(response.get("choices")[0].get("message"))
        return _message

    @staticmethod
    def parse_usage(response: dict):
        if not response.get("usage"):
            raise ValidationError("usage is empty")
        return response.get("usage").get("total_tokens")

    async def create(self,
                     **kwargs
                     ) -> OpenaiResult:
        """
        请求
        :return:
        """
        """
        curl https://api.openai.com/v1/completions \
        -H "Content-Type: application/json" \
        -H "Authorization: Bearer YOUR_API_KEY" \
        -d '{"model": "text-davinci-003", "prompt": "Say this is a test", "temperature": 0, "max_tokens": 7}'
        """
        # check
        num_tokens_from_messages = TokenizerObj.num_tokens_from_messages(self.messages,
                                                                         model=self.model,
                                                                         )
        if num_tokens_from_messages > self.get_token_limit(self.model):
            raise ValidationError("messages_box num_tokens > max_tokens")

        # Clear tokenizer encode cache
        # TokenizerObj.clear_cache()
        _data = self.dict(exclude_none=True, exclude={"config", "echo"})
        # 返回请求
        headers = {
            "User-Agent": "Mozilla/5.0",
            "Authorization": f"Bearer {self.config.api_key}",
            "api-key": f"{self.config.api_key}",
        }
        if self.config.org_id:
            headers["Openai-Organization"] = self.config.org_id
        try:
            _response = await request(
                method="POST",
                url=self.config.endpoint,
                data=_data,
                headers=headers,
                proxy=self.get_proxy_settings().proxy_address,
                json_body=True
            )
            return_result = OpenaiResult.parse_obj(_response)
        except httpx.ConnectError as e:
            logger.error(f"Openai connect error: {e}")
            raise e
        if self.echo:
            logger.info(f"Openai response: {return_result}")
        return return_result



# -*- coding: utf-8 -*-
# Source: https://github.com/langchain-ai/langchain/blob/master/libs/langchain/langchain/utilities/redis.py
from __future__ import annotations

import json

import redis
from loguru import logger
from pydantic import BaseSettings, Field, root_validator

from llmkira.sdk.schema import Message, parse_message_dict
from .utils import get_client


class RedisChatMessageHistory(object):
    class RedisSettings(BaseSettings):
        redis_url: str = Field("redis://localhost:6379/0", env="REDIS_DSN")
        redis_key_prefix: str = "llm_message_store_1:"

        class Config:
            env_file = '.env'
            env_file_encoding = 'utf-8'

        @root_validator
        def redis_is_connected(cls, values):
            redis_url = values.get("redis_url")
            try:
                get_client(redis_url=redis_url)
            except redis.exceptions.ConnectionError as error:
                logger.error(error)
                raise ValueError("Could not connect to Redis")
            else:
                logger.success("Redis connect success")
            return values

    def __init__(
            self,
            session_id: str,
            ttl: int,
            redis_config: RedisSettings = RedisSettings(),
    ):
        try:
            import redis
        except ImportError:
            raise ImportError(
                "Could not import redis python package. "
                "Please install it with `pip install redis`."
            )

        try:
            self.redis_client = get_client(redis_url=redis_config.redis_url)
        except redis.exceptions.ConnectionError as error:
            logger.error(error)

        self.session_id = session_id
        self.key_prefix = redis_config.redis_key_prefix
        self.ttl = ttl

    @property
    def key(self) -> str:
        """Construct the record key to use"""
        return self.key_prefix + self.session_id

    @property
    def messages(self) -> List[BaseMessage]:  # type: ignore
        """Retrieve the messages_box from Redis"""
        _items = self.redis_client.lrange(self.key, 0, -1)
        items = [json.loads(m.decode("utf-8")) for m in _items[::-1]]
        messages = [parse_message_dict(m) for m in items]
        return messages

    def add_message(self, message: Message) -> None:
        """Append the message to the record in Redis"""
        self.redis_client.lpush(self.key, message.json())
        if self.ttl:
            self.redis_client.expire(self.key, self.ttl)

    def clear(self) -> None:
        """Clear session memory from Redis"""
        self.redis_client.delete(self.key)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/21 下午9:56
# @Author  : sudoskys
# @File    : rss.py
# @Software: PyCharm


# 这里其实是一个推送的单端实例，主要展示了 Router 的用法，让不同平台的推送可以通过 Router 进行交换

import asyncio
import json
import socket

import feedparser
from inscriptis import get_text
from loguru import logger
from pydantic import BaseModel
from telebot import formatting
from telebot.formatting import escape_markdown

from llmkira.cache.redis import cache
from llmkira.middleware.router import RouterManager
from llmkira.middleware.router.schema import router_set

__sender__ = "rss"

from llmkira.task import Task, TaskHeader

router_set(role="sender", name=__sender__)


def sha1(string: str):
    import hashlib
    _sha1 = hashlib.sha1()
    _sha1.update(string.encode('utf-8'))
    return _sha1.hexdigest()[:10]


class RssApp(object):

    def parse_entry(self, entry):
        """
        {
                "title": entry["title"],
                "url": entry["link"],
                "id": entry["id"],
                "author": entry["author"],
                "summary": html2text.html2text(entry["summary"]),
            }
        """
        message = formatting.format_text(
            formatting.mbold(entry["title"]),
            "\n",
            escape_markdown(entry["summary"]),
            formatting.mlink(entry["author"], entry["url"]),
            separator="\n",
        )
        return message

    async def task(self):
        _router_list = RouterManager().get_router_by_sender(__sender__)
        for router in _router_list:
            try:
                logger.info(f"Sender:rss sub {router.rules} sent")
                _title, _entry = await Rss(feed_url=router.rules).get_updates()
                for item in _entry:
                    await Task(queue=router.to_).send_task(
                        task=TaskHeader.from_router(
                            from_=router.from_,
                            to_=router.to_,
                            user_id=router.user_id,
                            method=router.method,
                            message_text=self.parse_entry(item),
                        )
                    )
            except Exception as e:
                logger.error(e)
                _title, _entry = f"Error When receive sub {router.rules}", []
                await Task(queue=router.to_).send_task(
                    task=TaskHeader.from_router(
                        from_=router.from_,
                        to_=router.to_,
                        user_id=router.user_id,
                        method=router.method,
                        message_text=_title,
                    )
                )
                continue

    async def rss_polling(self, interval=60 * 60 * 1):
        logger.success("Sender Runtime:RSS Checker start")
        while True:
            # RSS 休眠
            await asyncio.sleep(interval)
            await self.task()


class Rss(object):
    """
    从缓存拉取，没有缓存就初始化，返回最后一条。
    有缓存则比对
    """

    class Update(BaseModel):
        title: str
        entry: dict

    def __init__(self, feed_url):
        # sha1
        self.db_key = f"rss:{sha1(feed_url)}"
        self.feed_url = feed_url

    def get_feed(self):
        socket.setdefaulttimeout(15)
        res = feedparser.parse(self.feed_url)
        try:
            json.dumps(res, indent=4, ensure_ascii=False)
        except Exception as e:
            raise ValueError(f"Fetch rss feed error {e},not valid rss json")
        entries = res["entries"]
        _title = res["feed"]["title"]
        _entry = {}
        for entry in entries:
            _entry[entry["id"]] = {
                "title": entry["title"],
                "url": entry["link"],
                "id": entry["id"],
                "author": entry["author"],
                "summary": get_text(entry["summary"]),
            }
        return self.Update(title=_title, entry=_entry)

    async def re_init(self, update: Update) -> (str, list):
        _entry = list(update.entry.values())[:1]
        await cache.set_data(key=self.db_key, value=update.json(), timeout=60 * 60 * 60 * 7)
        return update.title, _entry

    async def update(self, cache_, update_, keys):
        _return = []
        for key in keys:
            # copy
            cache_.entry[key] = update_.entry[key]
            _return.append(update_.entry[key])
        await cache.set_data(key=self.db_key, value=cache_.json(), timeout=60 * 60 * 60 * 7)
        return update_.title, _return

    async def get_updates(self):
        # 从缓存拉取
        _load = self.get_feed()
        _data = await cache.read_data(key=self.db_key)
        if not _data:
            return await self.re_init(_load)
        assert isinstance(_data, dict), "wrong rss data"
        _cache = self.Update.parse_obj(_data)

        # 验证是否全部不一样
        _old = list(_cache.entry.keys())
        _new = list(_load.entry.keys())
        _updates = [x for x in _new if x not in _old]

        # 全部不一样
        if len(_updates) == len(_new):
            return await self.re_init(_load)
        if len(_updates) == 0:
            return _load.title, []

        # 部分不一样
        return await self.update(cache_=_cache, update_=_load, keys=_updates)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/19 上午12:09
# @Author  : sudoskys
# @File    : sticker.py
# @Software: PyCharm
from io import BytesIO
from math import floor

from PIL import Image
from llmkira.middleware.chain_box import Chain, CHAIN_MANAGER
from llmkira.schema import RawMessage
from llmkira.sdk.endpoint.openai import Function
from llmkira.sdk.func_calling import BaseTool
from llmkira.sdk.func_calling.schema import FuncPair, PluginMetadata
from llmkira.task import Task, TaskHeader
from loguru import logger
from pydantic import validator, BaseModel

__plugin_name__ = "convert_to_sticker"
sticker = Function(name=__plugin_name__, description="Help user convert pictures to stickers")
sticker.add_property(
    property_name="yes_no",
    property_description="Is this run allowed (yes/no) If there a picture, please say yes",
    property_type="string",
    required=True
)
sticker.add_property(
    property_name="comment",
    property_description="thanks for this run",
    property_type="string",
    required=True
)


class Sticker(BaseModel):
    yes_no: str
    comment: str = "done"

    class Config:
        extra = "allow"

    @validator("yes_no")
    def delay_validator(cls, v):
        if v != "yes":
            v = "no"
        return v


async def resize_image(photo):
    logger.debug(f"Plugin:resize_image")
    image = Image.open(photo)
    if (image.width and image.height) < 512:
        size1 = image.width
        size2 = image.height
        if image.width > image.height:
            scale = 512 / size1
            size1new = 512
            size2new = size2 * scale
        else:
            scale = 512 / size2
            size1new = size1 * scale
            size2new = 512
        size1new = floor(size1new)
        size2new = floor(size2new)
        size_new = (size1new, size2new)
        image = image.resize(size_new)
    else:
        maxsize = (512, 512)
        image.thumbnail(maxsize)

    return image


class StickerTool(BaseTool):
    """
    搜索工具
    """
    function: Function = sticker
    keywords: list = ["转换", "贴纸", ".jpg", "图像", '图片']

    def pre_check(self):
        return True

    def func_message(self, message_text):
        """
        如果合格则返回message，否则返回None，表示不处理
        """
        for i in self.keywords:
            if i in message_text:
                return self.function
        # 正则匹配
        if self.pattern:
            match = self.pattern.match(message_text)
            if match:
                return self.function
        return None

    async def failed(self, platform, task, receiver, reason):
        try:
            await Task(queue=platform).send_task(
                task=TaskHeader(
                    sender=task.sender,
                    receiver=receiver,
                    task_meta=TaskHeader.Meta(
                        callback_forward=True,
                        callback=TaskHeader.Meta.Callback(
                            role="function",
                            name=__plugin_name__
                        ),
                    ),
                    message=[
                        RawMessage(
                            user_id=receiver.user_id,
                            chat_id=receiver.chat_id,
                            text="🍖 操作失败，原因：{}".format(reason)
                        )
                    ]
                )
            )
        except Exception as e:
            logger.error(e)

    async def callback(self, sign: str, task: TaskHeader):
        if sign == "reply":
            chain: Chain = await CHAIN_MANAGER.get_task(user_id=str(task.receiver.user_id))
            if chain:
                logger.info(f"{__plugin_name__}:chain callback locate in {sign} be sent")
                await Task(queue=chain.address).send_task(task=chain.arg)
            return True
        else:
            return False

    async def run(self, task: TaskHeader, receiver: TaskHeader.Location, arg, **kwargs):
        """
        处理message，返回message
        """
        try:
            _file = []
            for item in task.message:
                assert isinstance(item, RawMessage), "item must be RawMessage"
                if item.file:
                    for i in item.file:
                        _file.append(i)
            _set = Sticker.parse_obj(arg)
            _file_obj = [await RawMessage.download_file(file_id=i.file_id) for i in sorted(set(_file), key=_file.index)]
            # 去掉None
            _file_obj = [item for item in _file_obj if item]
            _result = []
            if not _file_obj:
                return
            for item in _file_obj:
                image = await resize_image(BytesIO(item.file_data))
                file = BytesIO()
                file.name = "sticker.webp"
                image.save(file, "WEBP")
                file.seek(0)
                file_obj = await RawMessage.upload_file(name="sticker.webp", data=file.getvalue())
                _result.append(file_obj)
            # META
            _meta = task.task_meta.child(__plugin_name__)
            _meta.callback_forward = True
            _meta.callback_forward_reprocess = False
            _meta.callback = TaskHeader.Meta.Callback(
                role="function",
                name=__plugin_name__
            )

            await Task(queue=receiver.platform).send_task(
                task=TaskHeader(
                    sender=task.sender,
                    receiver=receiver,
                    task_meta=_meta,
                    message=[
                        RawMessage(
                            user_id=receiver.user_id,
                            chat_id=receiver.chat_id,
                            file=_result,
                            text=_set.comment
                        )
                    ]
                )
            )

            logger.debug("convert_to_sticker say: {}".format(_set.yes_no))
        except Exception as e:
            logger.exception(e)
            await self.failed(platform=receiver.platform, task=task, receiver=receiver, reason=str(e))


__plugin_meta__ = PluginMetadata(
    name=__plugin_name__,
    description="Convert pictures to stickers",
    usage=str(StickerTool().keywords),
    openapi_version="20231013",
    function={
        FuncPair(function=sticker, tool=StickerTool)
    },
)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/21 下午10:19
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
from typing import List

from llmkira.cache.redis import cache
from llmkira.schema import singleton
from llmkira.utils import sync
from .schema import RouterCache, Router


@singleton
class RouterManager(object):
    def __init__(self):
        self.__redis_key__ = "router"
        self.router = sync(self._sync())

    async def _upload(self):
        assert isinstance(self.router, RouterCache), "router info error"
        self.router = RouterCache.parse_obj(self.router.dict())
        return await cache.set_data(key=self.__redis_key__, value=self.router.json())

    async def _sync(self) -> RouterCache:
        _cache = await cache.read_data(key=self.__redis_key__)
        if not _cache:
            return RouterCache()
        try:
            sub_info = RouterCache().parse_obj(_cache)
        except Exception:
            raise Exception(f"not found router info")
        return sub_info

    def get_router_by_user(self, to_: str, user_id: int) -> List[Router]:
        _all = self.router.router
        return [router for router in _all if router.user_id == user_id and router.to_ == to_]

    def get_router_by_sender(self, from_: str) -> List[Router]:
        _all = self.router.router
        return [router for router in _all if router.from_ == from_]

    def add_router(self, router: Router):
        self.router.router.append(router)
        return sync(self._upload())

    def remove_router(self, router: Router):
        if router in self.router.router:
            self.router.router.remove(router)
        return sync(self._upload())



# -*- coding: utf-8 -*-

import re


class OpenApiError(Exception):
    """插件错误基类。"""

    def __init__(self, message: str):
        super().__init__(message)


def escape_tag(s: str) -> str:
    """用于记录带颜色日志时转义 `<tag>` 类型特殊标签

    参考: [loguru color 标签](https://loguru.readthedocs.io/en/stable/api/logger.html#color)

    参数:
        s: 需要转义的字符串
    """
    return re.sub(r"</?((?:[fb]g\s)?[^<>\s]*)>", r"\\\g<0>", s)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/16 上午11:05
# @Author  : sudoskys
# @File    : action.py
# @Software: PyCharm
import hashlib
import json
from typing import List

import tiktoken
from pydantic import BaseModel

from llmkira.sdk.filter.evaluate import Sim
from llmkira.sdk.schema import Message, standardise


# 生成MD5
def generate_md5(string):
    hl = hashlib.md5()
    hl.update(string.encode(encoding='utf-8'))
    return str(hl.hexdigest())


class Tokenizer(object):
    __encode_cache = {}

    def clear_cache(self):
        self.__encode_cache = {}

    def num_tokens_from_messages(self, messages: List[Message], model="gpt-3.5-turbo-0613") -> int:
        """Return the number of tokens used by a list of messages_box."""
        try:
            encoding = tiktoken.encoding_for_model(model)
        except KeyError:
            print("Warning: model not found. Using cl100k_base encoding.")
            encoding = tiktoken.get_encoding("cl100k_base")
        if model in {
            "gpt-3.5-turbo-0613",
            "gpt-3.5-turbo-16k-0613",
            "gpt-4-0314",
            "gpt-4-32k-0314",
            "gpt-4-0613",
            "gpt-4-32k-0613",
        }:
            tokens_per_message = 3
            tokens_per_name = 1
        elif model == "gpt-3.5-turbo-0301":
            tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
            tokens_per_name = -1  # if there's a name, the role is omitted
        elif "gpt-3.5-turbo" in model:
            print("Warning: gpt-3.5-turbo may update over time. Returning num tokens assuming gpt-3.5-turbo-0613.")
            return self.num_tokens_from_messages(messages, model="gpt-3.5-turbo-0613")
        elif "gpt-4" in model:
            print("Warning: gpt-4 may update over time. Returning num tokens assuming gpt-4-0613.")
            return self.num_tokens_from_messages(messages, model="gpt-4-0613")
        else:
            raise NotImplementedError(
                f"""num_tokens_from_messages() is not implemented for model {model}."""
            )
        num_tokens = 0
        for message in messages:
            message = standardise(message)
            num_tokens += tokens_per_message
            for key, value in message.dict().items():
                if isinstance(value, dict):
                    value = json.dumps(value, ensure_ascii=False)
                _uid = generate_md5(str(value))
                # 缓存获取 cache，减少重复 encode 次数
                if _uid in self.__encode_cache:
                    _tokens = self.__encode_cache[_uid]
                else:
                    _tokens = len(encoding.encode(value))
                    self.__encode_cache[_uid] = _tokens
                num_tokens += _tokens
                if key == "name":
                    num_tokens += tokens_per_name
        num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens


TokenizerObj = Tokenizer()


class Scraper(BaseModel):
    """
    刮削器
    始终按照顺序排列，削除得分低的条目
    Scraper is a class that sorts a list of messages_box by their score.
    """

    class Sorter(BaseModel):
        message: Message
        # 得分
        score: float
        # 顺序
        order: int

    # 消息列表
    messages_box: List[Sorter] = []
    # 最大消息数
    max_messages: int = 12

    # 计数器
    tick: int = 0

    # 方法：添加消息
    def add_message(self, message: Message, score: float):
        if not hasattr(message, "content"):
            return None
        self.messages_box.append(self.Sorter(message=message, score=score, order=self.tick))
        self.tick += 1
        # 按照顺序排序
        self.messages_box.sort(key=lambda x: x.order)
        while len(self.messages_box) > self.max_messages:
            self.messages_box.pop(0)

    # 方法：获取消息
    def get_messages(self) -> List[Message]:
        # 按照顺序排序
        # self.messages_box.sort(key=lambda x: x.order)
        _message = [standardise(sorter.message) for sorter in self.messages_box]
        # 去重
        # [*dict.fromkeys(_message)]
        # -> unhashable type: 'Message'
        # logger.debug(_message)
        return _message

    def build_messages(self):
        # 只取三个，末位匹配
        _message = self.get_messages()
        if len(_message) < 3:
            return _message
        _build = []
        _must = _message[-3:]
        _check_list = _message[:-3]
        _match_sentence = _message[-1:][0].content
        for item_obj in _check_list:
            if Sim.cosion_similarity(pre=_match_sentence, aft=item_obj.content) < 0.9:
                _build.append(item_obj)
            else:
                pass
                # logger.warning(f"ignore sim item {item_obj}")
        _build.extend(_must)
        return _build

    # 方法：获取消息数
    def get_num_messages(self) -> int:
        return len(self.messages_box)

    # 方法：清除消息到负载
    def reduce_messages(self, limit: int = 2048):
        # 预留位
        if limit > 1000:
            limit = limit - 250
        else:
            limit = limit - 70

        # 执行删除操作
        if TokenizerObj.num_tokens_from_messages(self.get_messages()) > limit:
            # 从最旧开始删除
            self.messages_box.sort(key=lambda x: x.order)
            while TokenizerObj.num_tokens_from_messages(self.get_messages()) > limit:
                if len(self.messages_box) > 1:
                    self.messages_box.pop(0)
                else:
                    self.messages_box[0].message.content = self.messages_box[0].message.content[:limit]
        # 按照顺序排序
        self.messages_box.sort(key=lambda x: x.order)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/17 下午8:19
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm

from llmkira.middleware.chain_box import Chain, AUTH_MANAGER
from llmkira.middleware.router import RouterManager, Router
from llmkira.middleware.user import SubManager
from llmkira.schema import RawMessage
from llmkira.sdk.func_calling.register import ToolRegister
from llmkira.sdk.memory.redis import RedisChatMessageHistory
from llmkira.sender.telegram.utils import parse_command, is_valid_url
from llmkira.setting.telegram import BotSetting
from llmkira.task import Task, TaskHeader
from loguru import logger
from telebot import formatting, util
from telebot import types
from telebot.async_telebot import AsyncTeleBot
from telebot.asyncio_storage import StateMemoryStorage
from telebot.formatting import escape_markdown

StepCache = StateMemoryStorage()
__sender__ = "telegram"
TelegramTask = Task(queue=__sender__)
__default_function_enable__ = True


class TelegramBotRunner(object):
    def __init__(self):
        self.bot = None
        self.proxy = None

    async def telegram(self):
        if not BotSetting.available:
            logger.warning("Sender Runtime:TelegramBot Setting not available")
            return None
        self.bot = AsyncTeleBot(BotSetting.token, state_storage=StepCache)
        self.proxy = BotSetting.proxy_address
        bot = self.bot
        if self.proxy:
            from telebot import asyncio_helper
            asyncio_helper.proxy = self.proxy
            logger.info("TelegramBot proxy_tunnels being used!")

        def is_command(text, command, check_empty=False):
            if check_empty:
                if len(text.split(" ")) == 1:
                    return False
            if text.startswith(f"{command} ") or text.startswith(f"{command}@{BotSetting.bot_username}"):
                return True
            if text == command:
                return True
            return False

        def is_empty_command(text: str):
            if not text.startswith("/"):
                return False
            if len(text.split(" ")) == 1:
                return True
            return False

        async def is_admin(message: types.Message):
            _got = await bot.get_chat_member(message.chat.id, message.from_user.id)
            return _got.status in ['administrator', 'sender']

        async def auth_reloader(uuid, user_id):
            chain: Chain = await AUTH_MANAGER.get_auth(uuid=uuid, user_id=str(user_id))
            if chain:
                logger.info(f"Auth Task be sent\n--uuid {uuid} --user {user_id}")
                await Task(queue=chain.address).send_task(task=chain.arg)
                return True
            else:
                return False

        async def telegram_to_file(file):
            assert hasattr(file, "file_id"), "file_id not found"
            name = file.file_id
            _file_info = await bot.get_file(file.file_id)
            downloaded_file = await bot.download_file(_file_info.file_path)
            if isinstance(file, types.PhotoSize):
                name = f"{_file_info.file_unique_id}.jpg"
            if isinstance(file, types.Document):
                name = file.file_name
            # TODO 等待断点来规范化文件类型等消息,确保可以传入正常的文件ID
            return await RawMessage.upload_file(name=name, data=downloaded_file)

        async def create_task(message: types.Message, funtion_enable: bool = False):
            _file = []
            if message.text:
                message.text = message.text.lstrip("/chat").lstrip("/task")
            if message.photo:
                _file.append(await telegram_to_file(message.photo[-1]))
            if message.document:
                if message.document.file_size < 1024 * 1024 * 10:
                    _file.append(await telegram_to_file(message.document))
            if message.reply_to_message:
                if message.reply_to_message.photo:
                    _file.append(await telegram_to_file(message.reply_to_message.photo[-1]))
                if message.reply_to_message.document:
                    if message.reply_to_message.document.file_size < 1024 * 1024 * 10:
                        _file.append(await telegram_to_file(message.reply_to_message.document))
            message.text = message.text if message.text else ""
            logger.info(
                f"telegram:create task from {message.chat.id} {message.text[:300]} funtion_enable:{funtion_enable}")
            try:
                success, logs = await TelegramTask.send_task(
                    task=TaskHeader.from_telegram(
                        message,
                        file=_file,
                        task_meta=TaskHeader.Meta(function_enable=funtion_enable, sign_as=(0, "root", "telegram")),
                        trace_back_message=[message.reply_to_message]
                    ))
                if not success:
                    await bot.reply_to(message, text=logs)
            except Exception as e:
                logger.exception(e)

        @bot.message_handler(commands='clear_rset', chat_types=['private'])
        async def listen_clear_rset_command(message: types.Message):
            # _cmd, _arg = parse_command(command=message.text)
            _tips = "🪄 Done"
            await SubManager(user_id=f"{__sender__}:{message.from_user.id}").clear_endpoint()
            return await bot.reply_to(
                message,
                text=formatting.format_text(
                    formatting.mbold(_tips),
                    separator="\n"
                ),
                parse_mode="MarkdownV2"
            )

        @bot.message_handler(commands='rset_key', chat_types=['private'])
        async def listen_rset_key_command(message: types.Message):
            _cmd, _arg = parse_command(command=message.text)
            if not _arg:
                return
            if len(_arg) > 10:
                _tips = "🪄 Done"
                await SubManager(user_id=f"{__sender__}:{message.from_user.id}").set_endpoint(endpoint=_arg)
            else:
                _tips = f"🪄 {_arg} is too short"
            return await bot.reply_to(
                message,
                text=formatting.format_text(
                    formatting.mbold(_tips),
                    separator="\n"
                ),
                parse_mode="MarkdownV2"
            )

        @bot.message_handler(commands='rset_endpoint', chat_types=['private'])
        async def listen_rset_endpoint_command(message: types.Message):
            _cmd, _arg = parse_command(command=message.text)
            if not _arg:
                return
            if is_valid_url(_arg):
                _tips = "🪄 Done"
                await SubManager(user_id=f"{__sender__}:{message.from_user.id}").set_endpoint(endpoint=_arg)
            else:
                _tips = f"🪄 {_arg} is not a valid url"
            return await bot.reply_to(
                message,
                text=formatting.format_text(
                    formatting.mbold(_tips),
                    separator="\n"
                ),
                parse_mode="MarkdownV2"
            )

        @bot.message_handler(commands='bind', chat_types=['private'])
        async def listen_bind_command(message: types.Message):
            _cmd, _arg = parse_command(command=message.text)
            if not _arg:
                return
            _manager = RouterManager()
            try:
                router = Router.build_from_receiver(receiver=__sender__, user_id=message.from_user.id, dsn=_arg)
                _manager.add_router(router=router)
                router_list = _manager.get_router_by_user(user_id=message.from_user.id, to_=__sender__)
            except Exception as e:
                logger.exception(e)
                return await bot.reply_to(
                    message,
                    text=formatting.format_text(
                        formatting.mbold(str(e)),
                        separator="\n"
                    ),
                    parse_mode="MarkdownV2"
                )
            return await bot.reply_to(
                message,
                text=formatting.format_text(
                    formatting.mbold("🪄 Done"),
                    *[f"`{escape_markdown(item.dsn(user_dsn=True))}`" for item in router_list],
                    separator="\n"
                ),
                parse_mode="MarkdownV2"
            )

        @bot.message_handler(commands='unbind', chat_types=['private'])
        async def listen_unbind_command(message: types.Message):
            _cmd, _arg = parse_command(command=message.text)
            if not _arg:
                return None
            _manager = RouterManager()
            try:
                router = Router.build_from_receiver(receiver=__sender__, user_id=message.from_user.id, dsn=_arg)
                _manager.remove_router(router=router)
                router_list = _manager.get_router_by_user(user_id=message.from_user.id, to_=__sender__)
            except Exception as e:
                logger.exception(e)
                return await bot.reply_to(
                    message,
                    text=formatting.format_text(
                        formatting.mbold(str(e)),
                        separator="\n"
                    ),
                    parse_mode="MarkdownV2"
                )
            return await bot.reply_to(
                message,
                text=formatting.format_text(
                    formatting.mbold("🪄 Done"),
                    *[f"`{escape_markdown(item.dsn(user_dsn=True))}`" for item in router_list],
                    separator="\n"
                ),
                parse_mode="MarkdownV2"
            )

        @bot.message_handler(commands='clear', chat_types=['private', 'supergroup', 'group'])
        async def listen_help_command(message: types.Message):
            RedisChatMessageHistory(session_id=f"{__sender__}:{message.from_user.id}", ttl=60 * 60 * 1).clear()
            return await bot.reply_to(
                message,
                text=formatting.format_text(
                    formatting.mbold("🪄 Done"),
                    separator="\n"
                ),
                parse_mode="MarkdownV2"
            )

        @bot.message_handler(commands='help', chat_types=['private', 'supergroup', 'group'])
        async def listen_help_command(message: types.Message):
            from llmkira.sender.telegram.event import help_message
            _message = await bot.reply_to(
                message,
                text=formatting.format_text(
                    formatting.mbold("🥕 Help"),
                    escape_markdown(help_message()),
                    separator="\n"
                ),
                parse_mode="MarkdownV2"
            )

        @bot.message_handler(content_types=['text', 'photo', 'document'], chat_types=['private'])
        async def handle_private_msg(message: types.Message):
            """
            自动响应私聊消息
            """
            message.text = message.text if message.text else message.caption
            if not message.text:
                return None
            if is_command(text=message.text, command="/auth"):
                if not is_empty_command(text=message.text):
                    _cmd, _arg = parse_command(command=message.text)
                    try:
                        _run = await auth_reloader(uuid=_arg, user_id=message.from_user.id)
                    except Exception as e:
                        logger.error(f"[270563]auth_reloader failed {e}")
                        _run = None
                    if _run:
                        return await bot.reply_to(message, text="🪄 Auth Pass")
                    return await bot.reply_to(message,
                                              text="❌ Auth failed,You dont have permission or the task do not exist"
                                              )
            if is_command(text=message.text, command="/task"):
                return await create_task(message, funtion_enable=True)
            if is_command(text=message.text, command="/tool"):
                _tool = ToolRegister().functions
                _paper = [[c.name, c.description] for name, c in _tool.items()]
                arg = [
                    formatting.mbold(item[0]) +
                    formatting.mbold(" - ") +
                    formatting.mitalic(item[1])
                    for item in _paper
                ]
                return await bot.reply_to(
                    message,
                    text=formatting.format_text(
                        formatting.mbold("🔧 Tool List"),
                        *arg,
                        separator="\n"
                    ),
                    parse_mode="MarkdownV2"
                )
            return await create_task(message, funtion_enable=__default_function_enable__)

        @bot.message_handler(content_types=['text', 'photo', 'document'], chat_types=['supergroup', 'group'])
        async def handle_group_msg(message: types.Message):
            """
            自动响应群组消息
            """
            message.text = message.text if message.text else message.caption
            if not message.text:
                return None
            if is_command(text=message.text, command="/tool"):
                _tool = ToolRegister().functions
                _paper = [[c.name, c.description] for name, c in _tool.items()]
                arg = [
                    formatting.mbold(item[0]) +
                    formatting.mbold(" - ") +
                    formatting.mitalic(item[1])
                    for item in _paper
                ]
                return await bot.reply_to(
                    message,
                    text=formatting.format_text(
                        formatting.mbold("🔧 Tool List"),
                        *arg,
                        separator="\n"
                    ),
                    parse_mode="MarkdownV2"
                )
            if is_command(text=message.text, command="/auth"):
                if not is_empty_command(text=message.text):
                    _cmd, _arg = parse_command(command=message.text)
                    try:
                        _run = await auth_reloader(uuid=_arg, user_id=message.from_user.id)
                    except Exception as e:
                        logger.error(f"[270563]auth_reloader failed {e}")
                        _run = None
                    if _run:
                        return await bot.reply_to(message, text="🪄 Auth Pass")
                    return await bot.reply_to(message,
                                              text="❌ Auth failed,You dont have permission or the task do not exist"
                                              )
            if is_command(text=message.text, command="/chat"):
                if is_empty_command(text=message.text):
                    return await bot.reply_to(message, text="?")
                return await create_task(message, funtion_enable=__default_function_enable__)
            if is_command(text=message.text, command="/task"):
                if is_empty_command(text=message.text):
                    return await bot.reply_to(message, text="?")
                return await create_task(message, funtion_enable=True)
            if f"@{BotSetting.bot_username} " in message.text or message.text.endswith(f" @{BotSetting.bot_username}"):
                return await create_task(message, funtion_enable=__default_function_enable__)
            # 检查回复
            if message.reply_to_message:
                # 回复了 Bot
                if message.reply_to_message.from_user.id == BotSetting.bot_id:
                    return await create_task(message, funtion_enable=__default_function_enable__)

        from telebot import asyncio_filters
        bot.add_custom_filter(asyncio_filters.IsAdminFilter(bot))
        bot.add_custom_filter(asyncio_filters.ChatFilter())
        bot.add_custom_filter(asyncio_filters.StateFilter(bot))
        logger.success("Sender Runtime:TelegramBot start")
        await bot.infinity_polling(
            allowed_updates=util.update_types,
            skip_pending=True,
            timeout=20,
            request_timeout=20
        )



# -*- coding: utf-8 -*-
# @Time    : 2023/8/18 下午6:04
# @Author  : sudoskys
# @File    : __init__.py
# @Software: PyCharm
import os

from loguru import logger

pkg_path = os.path.dirname(__file__)
pkg_name = os.path.basename(pkg_path)

"""
import os
__all__ = [f.strip(".py") for f in os.listdir(os.path.abspath(os.path.dirname(__file__))) if
           f.endswith('.py') and "_" not in f]
"""


def setup():
    logger.info("====== Plugin: setup start ======")



# -*- coding: utf-8 -*-

import re
from abc import abstractmethod, ABC
from dataclasses import dataclass, field
from types import ModuleType
from typing import Optional, Type, Dict, Any, List, Union, Set, final

from pydantic import BaseModel, Field, validator

from ..schema import Function


class BaseTool(ABC, BaseModel):
    """
    基础工具类，所有工具类都应该继承此类
    """
    __slots__ = ()
    silent: bool = False
    function: Function
    keywords: List[str] = Field([], description="关键词")
    pattern: Optional[re.Pattern] = None
    require_auth: bool = False
    repeatable: bool = False

    # TODO 未来版本支持
    require_auth_kwargs: dict = {}

    @final
    @validator("keywords", pre=True)
    def check_keywords(cls, v):
        for i in v:
            if not isinstance(i, str):
                raise ValueError(f"keyword must be str, got {type(i)}")
            if len(i) > 20:
                raise ValueError(f"keyword must be less than 20 characters, got {len(i)}")
            if len(i) < 2:
                raise ValueError(f"keyword must be more than 2 characters, got {len(i)}")
        return v

    @abstractmethod
    def pre_check(self) -> Union[bool, str]:
        """
        预检查，如果不合格则返回False，合格则返回True
        返回字符串表示不合格，且有原因
        """
        return ...

    @abstractmethod
    def func_message(self, message_text):
        """
        如果合格则返回message，否则返回None，表示不处理
        """
        for i in self.keywords:
            if i in message_text:
                return self.function
        # 正则匹配
        if self.pattern:
            match = self.pattern.match(message_text)
            if match:
                return self.function
        return None

    @abstractmethod
    async def failed(self, platform, task, receiver, reason):
        """
        处理失败
        """
        return ...

    @abstractmethod
    async def run(self, task, receiver, arg, **kwargs):
        """
        处理message，返回message
        """
        return ...

    @abstractmethod
    async def callback(self, sign: str, task):
        """
        回调
        """
        return ...


@dataclass(eq=False)
class FuncPair:
    function: Function
    tool: Type[BaseTool]
    extra_arg: Dict[Any, Any] = field(default_factory=dict)

    @property
    def name(self):
        return self.function.name


@dataclass(eq=False)
class PluginMetadata:
    """
    插件元信息，由插件编写者提供
    """

    name: str
    """插件名称"""
    description: str
    """插件功能介绍"""
    usage: str
    """插件使用方法"""
    openapi_version: str
    """适应版本"""
    function: Set[FuncPair] = field(default_factory=set)
    """插件工具"""
    type: Optional[str] = None
    """插件类型"""
    homepage: Optional[str] = None
    """插件主页"""
    config: Optional[Type[BaseModel]] = None
    """插件配置项"""
    extra: Dict[Any, Any] = field(default_factory=dict)
    """插件额外信息，可由插件编写者自由扩展定义"""


@dataclass(eq=False)
class Plugin:
    """
    机制用途，不由用户实例化
    """

    name: str
    """插件索引标识，Bot 使用 文件/文件夹 名称作为标识符"""
    module: ModuleType
    """插件模块对象"""
    module_name: str
    """点分割模块路径"""
    manager: "PluginManager"
    """导入该插件的插件管理器"""
    parent_plugin: Optional["Plugin"] = None
    """父插件"""
    sub_plugins: Set["Plugin"] = field(default_factory=set)
    """子插件集合"""
    metadata: Optional[PluginMetadata] = None
    """插件元信息"""

# 探针插件加载机制 参考 nonebot2 项目设计



# -*- coding: utf-8 -*-
# @Time    : 2023/9/4 下午11:16
# @Author  : sudoskys
# @File    : evaluate.py
# @Software: PyCharm
import re
from typing import List

from loguru import logger

from .api.keyphrase.keyphrase import KeyPhraseExtraction
from .api.keywords.tfidf import TfidfKeywords
from .api.summarization.textrank_summarization import TextRankSummarization
from .api.summarization.tfidf_summarization import TfidfSummarization
from .api.text_similarity.cosion import CosionSimilarity
from .api.text_similarity.edit import EditSimilarity
from .api.text_similarity.simhash import SimHashSimilarity


class DetectSentence(object):
    """
    检测句子
    """

    @staticmethod
    def detect_language(sentence: str):
        """
        Detect language
        :param sentence: sentence
        :return: 两位大写语言代码 (EN, ZH, JA, KO, FR, DE, ES, ....)
        """
        # 如果全部是空格
        if sentence.isspace() or not sentence:
            return ""
        try:
            from . import langdetect_fasttext
            lang_type = langdetect_fasttext.detect(text=sentence.replace("\n", "").replace("\r", ""),
                                                   low_memory=True).get("lang").upper()
        except Exception as e:
            # handle error
            logger.trace(e)
            from . import langdetect_unicode
            lang_type = langdetect_unicode.detect(text=sentence.replace("\n", "").replace("\r", ""))[0][0].upper()
        return lang_type

    @staticmethod
    def detect_help(sentence: str) -> bool:
        """
        检测是否是包含帮助要求，如果是，返回True，否则返回False
        """
        _check = ['怎么做', 'How', 'how', 'what', 'What', 'Why', 'why', '复述', '复读', '要求你', '原样', '例子',
                  '解释', 'exp', '推荐', '说出', '写出', '如何实现', '代码', '写', 'give', 'Give',
                  '请把', '请给', '请写', 'help', 'Help', '写一', 'code', '如何做', '帮我', '帮助我', '请给我', '什么',
                  '为何', '给建议', '给我', '给我一些', '请教', '建议', '怎样', '如何', '怎么样',
                  '为什么', '帮朋友', '怎么', '需要什么', '注意什么', '怎么办', '助け', '何を', 'なぜ', '教えて',
                  '提案',
                  '何が', '何に',
                  '何をす', '怎麼做', '複述', '復讀', '原樣', '解釋', '推薦', '說出', '寫出', '如何實現', '代碼', '寫',
                  '請把', '請給', '請寫', '寫一', '幫我', '幫助我', '請給我', '什麼', '為何', '給建議', '給我',
                  '給我一些', '請教', '建議', '步驟', '怎樣', '怎麼樣', '為什麼', '幫朋友', '怎麼', '需要什麼',
                  '註意什麼', '怎麼辦']
        for item in _check:
            if item in sentence:
                return True
        return False

    @staticmethod
    def detect_code(sentence) -> bool:
        """
        Detect code，if code return True，else return False
        :param sentence: sentence
        :return: bool
        """
        code = False
        _reco = [
            '("',
            '")',
            ").",
            "()",
            "!=",
            "==",
        ]
        _t = len(_reco)
        _r = 0
        for i in _reco:
            if i in sentence:
                _r += 1
        if _r > _t / 2:
            code = True
        rms = [
            "print_r(",
            "var_dump(",
            'NSLog( @',
            'println(',
            '.log(',
            'print(',
            'printf(',
            'WriteLine(',
            '.Println(',
            '.Write(',
            'alert(',
            'echo(',
        ]
        for i in rms:
            if i in sentence:
                code = True
        return code


class Sim(object):
    """
    文本相似度计算，基于基础 jieba 分词 101 向量
    """

    @staticmethod
    def cosion_similarity(pre, aft):
        """
        基于余弦计算文本相似性 0 - 1 (1为最相似)
        :return: 余弦值
        """
        _cos = CosionSimilarity()
        _sim = _cos.similarity(pre, aft)
        return _sim

    @staticmethod
    def edit_similarity(pre, aft):
        """
        基于编辑计算文本相似性
        :return: 差距
        """
        _cos = EditSimilarity()
        _sim = _cos.edit_dist(pre, aft)
        return _sim

    @staticmethod
    def simhash_similarity(pre, aft):
        """
        采用simhash计算文本之间的相似性
        :return:
        """
        simhash = SimHashSimilarity()
        sim = simhash.run_simhash(pre, aft)
        # print("simhash result: {}\n".format(sim))
        return sim

    @staticmethod
    def vector_similarity(text1_vector: list, text2_vector: list):
        """
        计算两个向量的余弦相似度
        :param text1_vector: 向量1
        :param text2_vector: 向量2
        :return: 余弦相似度
        """
        from sklearn.metrics.pairwise import cosine_similarity
        return cosine_similarity([text1_vector], [text2_vector])[0][0]


class Cut(object):
    @staticmethod
    def english_sentence_cut(text) -> list:
        list_ = list()
        for s_str in text.split('.'):
            if '?' in s_str:
                list_.extend(s_str.split('?'))
            elif '!' in s_str:
                list_.extend(s_str.split('!'))
            else:
                list_.append(s_str)
        return list_

    @staticmethod
    def chinese_sentence_cut(text) -> list:
        """
        中文断句
        """
        text = re.sub('([。！？\?])([^’”])', r'\1\n\2', text)
        # 普通断句符号且后面没有引号
        text = re.sub('(\.{6})([^’”])', r'\1\n\2', text)
        # 英文省略号且后面没有引号
        text = re.sub('(\…{2})([^’”])', r'\1\n\2', text)
        # 中文省略号且后面没有引号
        text = re.sub('([.。！？\?\.{6}\…{2}][’”])([^’”])', r'\1\n\2', text)
        # 断句号+引号且后面没有引号
        return text.split("\n")

    def cut_chinese_sentence(self, text):
        """
        中文断句
        """
        p = re.compile("“.*?”")
        listr = []
        index = 0
        for i in p.finditer(text):
            temp = ''
            start = i.start()
            end = i.end()
            for j in range(index, start):
                temp += text[j]
            if temp != '':
                temp_list = self.chinese_sentence_cut(temp)
                listr += temp_list
            temp = ''
            for k in range(start, end):
                temp += text[k]
            if temp != ' ':
                listr.append(temp)
            index = end
        return listr

    def cut_sentence(self, sentence: str) -> List[str]:
        """
        分句
        :param sentence:
        :return:
        """
        language = DetectSentence.detect_language(sentence)
        if language == "CN":
            _reply_list = self.cut_chinese_sentence(sentence)
        elif language == "EN":
            # from nltk.tokenize import sent_tokenize
            _reply_list = self.english_sentence_cut(sentence)
        else:
            _reply_list = [sentence]
        if len(_reply_list) < 1:
            return [sentence]
        return _reply_list


class Extraction(object):
    @staticmethod
    def key_phrase_extraction(sentence: str):
        """
        关键词抽取
        """
        return KeyPhraseExtraction().key_phrase_extraction(text=sentence)

    @staticmethod
    def textrank_summarization(sentence: str, ratio=0.2):
        """
        采用 textrank 进行摘要抽取
        :param sentence: 待处理语句
        :param ratio: 摘要占文本长度的比例
        :return:
        """
        _sum = TextRankSummarization(ratio=ratio)
        _sum = _sum.analysis(sentence)
        return _sum

    @staticmethod
    def tfidf_summarization(sentence: str, ratio=0.5):
        """
        采用tfidf进行摘要抽取
        :param sentence:
        :param ratio: 摘要占文本长度的比例
        :return:
        """
        _sum = TfidfSummarization(ratio=ratio)
        _sum = _sum.analysis(sentence)
        return _sum

    @staticmethod
    def tfidf_keywords(keywords, delete_stopwords=True, top_k=5, with_weight=False):
        """
        tfidf 提取关键词
        :param keywords:
        :param delete_stopwords: 是否删除停用词
        :param top_k: 输出关键词个数
        :param with_weight: 是否输出权重
        :return: [(word, weight), (word1, weight1)]
        """
        tfidf = TfidfKeywords(delete_stopwords=delete_stopwords, topK=top_k, withWeight=with_weight)
        return tfidf.keywords(keywords)



# -*- coding: utf-8 -*-


import re

import cjieba
import numpy as np

from ..solo import singleton


# from wordcloud import WordCloud

@singleton
class KeyPhraseExtraction(object):
    def __init__(self, topk=50, method='tfidf', with_word=True):
        """
        :param topk: 根据前多少关键词生成短语
        :param method: tfidf / textrank
        :param with_word: 最后输出结果是否包含关键词
        """
        self.topk = topk
        self.method = method
        self.with_word = with_word

    def cut_sentences(self, text):
        """文本分句，然后分词"""
        sentences = re.findall(".*?[。？！]", text)
        cut_sentences = [cjieba.lcut(sent) for sent in sentences]
        return cut_sentences

    def key_words_extraction(self, text):
        """提取关键词"""
        keywords_score = []
        if self.method == 'tfidf':
            # keywords_score = jieba.analyse.extract_tags(text, topK=self.topk, withWeight=True)
            keywords_score = cjieba.extract(text, top_k=self.topk, with_weight=True)
        elif self.method == 'textrank':
            raise NotImplementedError('textrank method is not implemented')
            # keywords_score = jieba.analyse.textrank(text, topK=self.topk, withWeight=True)
        return {word: score for word, score in keywords_score}

    def key_phrase_extraction(self, text):
        keyword_score = self.key_words_extraction(text)
        keywords = keyword_score.keys()
        cut_sentences = self.cut_sentences(text)
        # print(keywords)
        # 将相邻的关键词进行拼接
        key_phrase = []
        for sent in cut_sentences:
            temp = []
            for word in sent:
                if word in keywords:
                    temp.append(word)
                else:
                    if len(temp) > 1:
                        if temp not in key_phrase:
                            key_phrase.append(temp)
                    temp = []

        # 短语之间可能存在冗余信息，进行过滤
        key_phrase_filter = []
        for phrase in key_phrase:
            flag = False
            for item in key_phrase_filter:
                if len(set(phrase) & set(item)) >= min(len(set(phrase)), len(set(item))) / 2.0:
                    flag = True
                    break
            if not flag:
                key_phrase_filter.append(phrase)

        # 给短语赋值权重, 设置短语最多包含三个关键词
        keyphrase_weight = {''.join(phrase[-3:]): np.mean([keyword_score[word] for word in phrase[-3:]])
                            for phrase in key_phrase_filter}

        if self.with_word:
            key_phrase_str = '|'.join(keyphrase_weight)
            for word, weight in keyword_score.items():
                if word not in key_phrase_str:
                    keyphrase_weight[word] = weight
        keyphrase_weight = dict(sorted(keyphrase_weight.items(), key=lambda x: x[1], reverse=True)[:self.topk])

        return keyphrase_weight


if __name__ == '__main__':
    with open('./data/test', encoding='utf-8') as f:
        text = f.read()
    key_phrase_extractor = KeyPhraseExtraction(topk=20)
    key_phrase = key_phrase_extractor.key_phrase_extraction(text)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/18 上午9:37
# @Author  : sudoskys
# @File    : llm_task.py
# @Software: PyCharm
import os
import time
from typing import List, Literal

from llmkira.middleware.user import SubManager, UserInfo
from llmkira.schema import RawMessage
from llmkira.sdk.endpoint import openai
from llmkira.sdk.endpoint.openai import Message
from llmkira.sdk.endpoint.openai.action import Scraper
from llmkira.sdk.memory.redis import RedisChatMessageHistory
from llmkira.task import TaskHeader
from llmkira.utils import sync
from loguru import logger

RULE = """[ASSISTANT RULE]
* RE-SENDING THE FUNCTION WITH THE SAME PARAMETERS IS NOT ALLOWED
* SPEAK IN A MORE HUMAN-LIKE MANNER
[ASSISTANT RULE]"""


class OpenaiMiddleware(object):
    """
    Openai中间件，用于处理消息转换和调用工具
    任务数据>转换器+函数填充>提取历史>放进刮削器>任务数据+刮削结果请求>获取Openai返回>进行声明通知/返回消息
    """

    def __init__(self, task: TaskHeader, function: List[openai.Function] = None):
        self.scraper = Scraper()  # 刮削器
        self.functions: List[openai.Function] = function
        if self.functions is None:
            self.functions = []
        self.task = task
        self.sub_manager = SubManager(user_id=f"{task.sender.platform}:{task.sender.user_id}")  # 由发送人承担接受者的成本
        self.message_history = RedisChatMessageHistory(
            session_id=f"{task.receiver.platform}:{task.receiver.user_id}",
            ttl=60 * 60 * 1
        )

    def build(self, auto_write_back, default_system: bool = True):
        """
           构建消息和函数列表
           :param auto_write_back: 自动回写缓存消息到数据库
              :param default_system: 是否添加默认系统时钟/环境消息
        """
        self.create_message(write_back=auto_write_back, default_system=default_system)
        self.append_function()
        return self

    def write_back(self,
                   message_list: List[RawMessage],
                   name: str = None,
                   role: Literal["user", "system", "function", "assistant"] = "user"):
        """
        写回消息到 Redis 数据库中
        function 写回必须指定 name
        """
        for message in message_list:
            self.message_history.add_message(message=Message(role=role, name=name, content=message.text))

    def append_function(self):
        """
        添加函数
        """
        # 去重
        _dict = {}
        for function in self.functions:
            _dict[function.name] = function
        self.functions = list(_dict.values())
        # 取消
        _ignore = sync(self.sub_manager.get_lock_plugin())
        for function in self.functions:
            if function.name in _ignore:
                self.functions.remove(function)

    def create_message(self, write_back=True, default_system: bool = True):
        """
        从人类消息和历史消息中构建请求所用消息
        """
        if default_system:
            localtime = time.asctime(time.localtime(time.time()))
            rule = f"<Time|{localtime}>" + RULE
            self.scraper.add_message(Message(role="system", name="system",
                                             content=rule),
                                     score=0)
        _history = []
        history_messages = self.message_history.messages
        for i, message in enumerate(history_messages):
            _history.append(message)
        # 刮削器合并消息，这里评价简写了。
        for i, _msg in enumerate(_history):
            self.scraper.add_message(_msg, score=len(str(_msg)))

        # 处理 人类 发送的消息
        if write_back:
            _buffer = []
            raw_message = self.task.message
            raw_message: List[RawMessage]
            for i, message in enumerate(raw_message):
                _buffer.append(Message(role="user", name=None, content=message.text))
            # 装样子添加评分
            # TODO 评分
            for i, _msg in enumerate(_buffer):
                self.scraper.add_message(_msg, score=len(str(_msg)) + 50)
            # 处理缓存后默认写回 Redis 数据库
            for _msg in _buffer:
                self.message_history.add_message(message=_msg)

    async def request_openai(self, disable_function: bool = False) -> openai.OpenaiResult:
        """
        处理消息转换和调用工具
        :param disable_function: 禁用函数
        :return:
        """
        # 模型内容
        model_name = os.getenv("OPENAI_API_MODEL", "gpt-3.5-turbo-0613")
        self.scraper.reduce_messages(limit=openai.Openai.get_token_limit(model=model_name))
        message = self.scraper.get_messages()
        assert message, "llm_task.py:message is None"
        _functions = self.functions if self.functions else None
        driver = self.sub_manager.llm_driver
        assert isinstance(driver, openai.Openai.Driver), "llm_task.py:driver type error"

        # 消息缓存读取和转换
        # 断点
        logger.info(f"[x] Openai request \n--message {message} \n--url {driver.endpoint} \n--org {driver.org_id} ")
        if disable_function or not _functions:
            logger.debug(f"[x] Openai function empty warn \n--disable function:{disable_function}")
        if disable_function:
            _functions = None
        endpoint = openai.Openai(
            config=driver,
            model=model_name,
            # presence_penalty=0.5,
            # frequency_penalty=0.3,
            messages=message,
            functions=_functions,
            echo=False
        )

        # 调用Openai
        result: openai.OpenaiResult = await endpoint.create()
        _message = result.default_message
        _usage = result.usage.total_tokens
        self.message_history.add_message(message=_message)
        await self.sub_manager.add_cost(
            cost=UserInfo.Cost(token_usage=_usage, token_uuid=driver.uuid, model_name=model_name)
        )
        logger.success(f"[x] Openai result \n--message {result.choices} \n--token {result.usage}")
        return result



# -*- coding: utf-8 -*-
# @Time    : 2023/7/10 下午9:43
# @Author  : sudoskys
# @File    : base.py
# @Software: PyCharm
from abc import abstractmethod, ABC
from typing import Any

PREFIX = 'base_bot:'


class AbstractDataClass(ABC):

    @abstractmethod
    async def ping(self):
        return True

    @abstractmethod
    def update_backend(self, backend):
        pass

    @abstractmethod
    async def set_data(self, key: str, value: Any, **kwargs) -> Any:
        pass

    @abstractmethod
    async def read_data(self, key: str) -> Any:
        pass




# -*- coding: utf-8 -*-
# @Time    : 2023/8/21 下午10:19
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
from typing import Literal, List

from pydantic import BaseModel, Field, validator

SENDER = {}
RECEIVER = {}


def router_set(role: Literal["sender", "receiver"], name: str):
    if role == "sender":
        SENDER[name] = name
    elif role == "receiver":
        RECEIVER[name] = name
    else:
        raise ValueError(f"role must in sender or receiver, not {role}")


ROUTER_METHOD = Literal["push", "chat", "task"]
ALLOW_METHOD = ["push", "chat", "task"]


class Router(BaseModel):
    from_: str = Field(..., description="Sender")
    to_: str = Field(..., description="Receiver")
    user_id: int = Field(..., description="用户ID")
    rules: str = Field(..., description="URL")
    method: ROUTER_METHOD = Field('push', description="")  # "summary"

    @classmethod
    def build_from_receiver(cls, receiver, user_id, dsn: str):
        try:
            from_, rules, _method = dsn.split("@", maxsplit=3)
        except Exception as e:
            raise ValueError(f"dsn error {dsn},{e}. exp: rss@http://rss.toml@push  available: {SENDER}")
        if from_ not in SENDER:
            raise ValueError(f"sender must in {SENDER}, not {from_}")
        if _method not in ALLOW_METHOD:
            raise ValueError(f"method must in {ALLOW_METHOD}, not {_method}")
        return cls(from_=from_, to_=receiver, user_id=user_id, rules=rules)

    def dsn(self, user_dsn=False):
        if user_dsn:
            return f"{self.from_}@{self.rules}@{self.method}"
        return f"{self.from_}@{self.to_}@{self.user_id}@{self.rules}@{self.method}"


class RouterCache(BaseModel):
    router: List[Router] = []

    @validator("router", always=True)
    def router_validate(cls, v):
        _dict = {}
        for item in v:
            _dict[item.dsn()] = item
        v = list(_dict.values())
        return v



# -*- coding: utf-8 -*-
# @Time    : 2023/8/18 下午8:34
# @Author  : sudoskys
# @File    : aps.py
# @Software: PyCharm
from apscheduler.executors.pool import ThreadPoolExecutor
from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from loguru import logger

jobstores = {
    'default': SQLAlchemyJobStore(url='sqlite:///jobs.sqlite')
}
executors = {
    'default': ThreadPoolExecutor(20)
}
job_defaults = {
    'coalesce': False,
    'max_instances': 3
}
from pytz import utc

SCHEDULER = AsyncIOScheduler(job_defaults=job_defaults, timezone=utc)


async def aps_start():
    logger.success("Receiver Runtime:APS Timer start")
    SCHEDULER.start()



# -*- coding: utf-8 -*-
# @Time    : 2023/9/4 下午11:24
# @Author  : sudoskys
# @File    : sublimate.py
# @Software: PyCharm
import re
from typing import List

import numpy as np
from pydantic import BaseModel

from .evaluate import Cut, Extraction, Sim

# 六边形评测，按照排名分配初始分数。
# 采用切分分裂条目，过滤无意义语句。
# 采用关键词筛选清洗 ，挑选。
# 相似度排序，装箱。
BREAK_SHORT = [
    "訂閱", "订阅", "播放量", "点赞",
    "相关视频", "重试", "举报", "内容合作", "ICP备", "公网安备", "哔哩哔哩bilibili_",
    "自动连播", "抖音短视频", "短视频", "更多精彩", "只看楼主", "上一篇", "#热议#",
    "关注问题", "🔞", "下载APP", "违法有害信息"
]


class Order(BaseModel):
    text: str
    x_order: int = -1
    y_score: int = -1
    z_sim: int = -1


class Sublimate(object):

    def __init__(self, sentences: List[str]):
        self.origin = sentences
        self.x_factor = 9  # 初始顺序
        self.y_factor = 3  # 主题得分
        self.z_factor = 1  # 相似度检查
        self.valuate: List[Order] = []
        if not self.origin:
            raise ValueError("Bad Arg For Sublimate")

    @staticmethod
    def _count_in(key_list: List[str], target: str):
        _count = 0
        for item in key_list:
            if item in target:
                _count = _count + 1
        return _count

    @staticmethod
    def real_len(string):
        _r = re.findall(r"[^Wd]+", string)
        return len("".join(_r))

    def wipe_sentence(self, min_limit: int, must_contain: List[str] = None):
        _wipe = self.valuate.copy()
        _new_copy = []
        for item_obj in _wipe:
            if self.real_len(item_obj.text) < min_limit:
                continue
            if self._count_in(BREAK_SHORT, item_obj.text) != 0:
                continue
            if must_contain:
                if self._count_in(must_contain, item_obj.text) != len(must_contain):
                    continue
            _new_copy.append(item_obj)
        #
        self.valuate: List[Order] = _new_copy

    def valuation(self, match_sentence: str, match_keywords: List[str] = None, min_limit: int = 13):
        # 分割并赋予位置
        self.valuate: List[Order] = []
        for index, item in enumerate(self.origin):
            _child = [item]
            if index > 3 and len(item) > 20:
                _child = Cut().cut_sentence(item)
            __score = ((1 - (index / len(self.origin))) * 100) * self.x_factor  # 评估计算 order,乘影响因子
            for child_item in _child:
                self.valuate.append(Order(text=child_item, x_order=__score))

        # 清洗数据
        self.wipe_sentence(min_limit=min_limit, must_contain=match_keywords)
        if not self.valuate:
            return []

        # 提取句子成分
        _keywords = Extraction.tfidf_keywords(match_sentence)  # 高开销注意
        # 对其筛选评分
        __total = len(_keywords)
        for order_obj in self.valuate:
            # 为每个条目赋予主题评分
            __score = 0
            for key in _keywords:
                if key in order_obj.text:
                    __score = __score + 1
            # 计算当前条目分数
            _y_score = ((__score / __total) * 100) * self.y_factor
            order_obj.y_score = _y_score

        # 相似度计算方法
        for order_obj in self.valuate:
            # 高开销注意
            _z_sim = ((Sim.cosion_similarity(pre=match_sentence, aft=order_obj.text)) * 100) * self.z_factor
            order_obj.z_sim = _z_sim

        # 计算
        # 初始化点
        origin = np.array((0, 0, 0))
        _result = {}
        for item_obj in self.valuate:
            _target = np.array((item_obj.x_order, item_obj.y_score, item_obj.z_sim))
            _result[item_obj.text] = float(np.linalg.norm(origin - _target))
        return sorted(_result.items(), reverse=True, key=lambda x: x[1])



import asyncio
import atexit
import json
from typing import Any

import httpx
from loguru import logger

from .error import RateLimitError, ServiceUnavailableError, AuthenticationError

__session_pool = {}


def llm_error_handler(code, message: str):
    if code == 429:
        raise RateLimitError(message)
    elif code == 500 or code != 401:
        raise ServiceUnavailableError(message)
    else:
        raise AuthenticationError(message)


def check_json_response(status_code, req_data: dict):
    """
    检查json响应
    """
    # 错误消息
    if req_data.get("error", None):
        _error = req_data.get("error", None)
        raise Exception(f"{status_code}:{_error.get('type')}:{_error.get('message')}")
    # 错误消息
    if isinstance(req_data.get("message"), str):
        raise Exception(f"{status_code}:{req_data.get('message')}")


async def request(
        method: str,
        url: str,
        params: dict = None,
        data: Any = None,
        headers: dict = None,
        json_body: bool = False,
        proxy: str = "",
        call_func=None,
        **kwargs,
):
    """
    请求
    :param call_func: 错误回调函数
    :param method:
    :param url:
    :param params:
    :param data:
    :param headers:
    :param json_body:
    :param proxy:
    :param kwargs: 参数
    :return:
    """
    if headers is None:
        headers = {}
    if params is None:
        params = {}
    config = {
        "method": method.upper(),
        "url": url,
        "params": params,
        "data": data,
        "headers": headers,
    }
    # 更新
    config.update(kwargs)
    if json_body:
        config["headers"]["Content-Type"] = "application/json"
        config["data"] = json.dumps(config["data"]).encode()
    # SSL
    # config["ssl"] = False

    # 请求
    session = get_session(proxy=proxy)
    resp = await session.request(**config)

    # 检查响应头 Content-Length
    content_length = resp.headers.get("content-length")
    if content_length and int(content_length) == 0:
        return None

    # 检查响应头 Content-Type
    content_type = resp.headers.get("content-type")
    if content_type.lower().find("application/json") == -1:
        logger.error(f"[78652]Server send a invalid response {resp.text}")
        raise Exception("Server send a invalid response with content-type:{}".format(content_type))

    try:
        req_data: dict = json.loads(resp.text)
    except json.JSONDecodeError:
        if resp.status_code != 200:
            raise Exception(f"Request {resp.status_code}:Request failed without any error message")
        raise Exception("Server send a invalid json response")
    try:
        check_json_response(resp.status_code, req_data)
    except Exception as e:
        # Dict Message 格式校验失败
        if call_func:
            call_func(req_data, headers)
        _status = req_data.get("status", None)
        llm_error_handler(
            code=_status,
            message=f"{resp.status_code}:{req_data.get('message')}"
        )
    return req_data


def get_session(proxy: str = ""):
    global __session_pool
    loop = asyncio.get_event_loop()
    session = __session_pool.get(loop, None)
    if session is None:
        if proxy:
            proxies = {"all://": proxy}
            session = httpx.AsyncClient(timeout=300, proxies=proxies)
        else:
            session = httpx.AsyncClient(timeout=300)
        __session_pool[loop] = session
    return session


@atexit.register
def __clean():
    """
    程序退出清理操作。
    """
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        return

    async def __clean_task():
        await __session_pool[loop].close()

    if loop.is_closed():
        loop.run_until_complete(__clean_task())
    else:
        loop.create_task(__clean_task())



# -*- coding: utf-8 -*-
# @Time    : 2023/8/21 上午12:06
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
from typing import List, Union, Optional

from llmkira.sdk.endpoint.openai import Openai
from pydantic import BaseModel, Field, BaseSettings, validator


# 负责任务前后的APIKey信息管理和验证机制(认证可信平台) 平台:ID


class UserInfo(BaseSettings):
    class Plugin(BaseModel):
        lock: List[str] = []

    class Cost(BaseModel):
        cost_by: str = Field("chat", description="环节")
        token_usage: int = Field(0)
        token_uuid: str = Field(None, description="API KEY 的 hash")
        model_name: str = Field(None, description="模型")

        @classmethod
        def by_function(cls, function_name: str,
                        token_usage: int,
                        token_uuid: str,
                        model_name: str,
                        ):
            return cls(cost_by=function_name, token_usage=token_usage, token_uuid=token_uuid, model_name=model_name)

    user_id: Union[str, int] = Field(None, description="用户ID")
    plugin_subs: Plugin = Field(Plugin(), description="插件的设置")
    costs: List[Cost] = Field([], description="消费记录")
    llm_driver: Optional[Openai.Driver] = None
    update_driver: bool = False

    def total_cost(self):
        return sum([cost.token_usage for cost in self.costs])

    @validator("user_id")
    def check_user_id(cls, v):
        return str(v)



# -*- coding: utf-8 -*-
# @Time    : 2023/9/24 上午12:48
# @Author  : sudoskys
# @File    : translate_doc.py
# @Software: PyCharm
import asyncio
import os
from io import BytesIO
from typing import List

from llmkira.middleware.chain_box import Chain, CHAIN_MANAGER
from llmkira.middleware.user import UserInfo, SubManager
from llmkira.schema import RawMessage
from llmkira.sdk.endpoint import openai
from llmkira.sdk.endpoint.openai import Function
from llmkira.sdk.func_calling import BaseTool, PluginMetadata
from llmkira.sdk.func_calling.schema import FuncPair
from llmkira.sdk.schema import Message, File
from llmkira.task import Task, TaskHeader
from loguru import logger
from pydantic import BaseModel
from tenacity import retry, stop_after_attempt, stop_after_delay, wait_fixed

__plugin_name__ = "translate_file"
translate = Function(name=__plugin_name__, description="Help user translate [ReadableFile],only support txt/md")
translate.add_property(
    property_name="language",
    property_description="What language should the text be translated into?",
    property_type="string",
    required=True
)
translate.add_property(
    property_name="file_id",
    property_description="regex: file_id=([a-z0-9]{8}), require ReadableFile[...]",
    property_type="string",
    required=True
)


class Translate(BaseModel):
    language: str
    file_id: str

    class Config:
        extra = "allow"


class TranslateTool(BaseTool):
    """
    搜索工具
    """
    function: Function = translate
    keywords: list = ["translate", "转换", "convert", "翻译", "译", "md", 'txt']

    def pre_check(self):
        try:
            import nltk
            nltk.download('punkt')
            nltk.download('averaged_perceptron_tagger')
            from unstructured.partition.auto import partition
        except Exception as e:
            logger.error(f"plugin:translate_doc:{e},pls check !pip install unstructured")
            return False
        return True

    def func_message(self, message_text):
        """
        如果合格则返回message，否则返回None，表示不处理
        """
        for i in self.keywords:
            if i in message_text:
                return self.function
        # 正则匹配
        if self.pattern:
            match = self.pattern.match(message_text)
            if match:
                return self.function
        return None

    async def failed(self, platform: str, task: TaskHeader, receiver: TaskHeader.Location, reason: str):
        try:
            await Task(queue=platform).send_task(
                task=TaskHeader(
                    sender=task.sender,
                    receiver=receiver,
                    task_meta=TaskHeader.Meta(
                        callback_forward=True,
                        callback=TaskHeader.Meta.Callback(
                            role="function",
                            name=__plugin_name__
                        ),
                    ),
                    message=[
                        RawMessage(
                            user_id=receiver.user_id,
                            chat_id=receiver.chat_id,
                            text="🍖 操作失败，原因：{}".format(reason)
                        )
                    ]
                )
            )
        except Exception as e:
            logger.error(e)

    @staticmethod
    @retry(stop=(stop_after_attempt(3) | stop_after_delay(10)), wait=wait_fixed(2), reraise=True)
    async def llm_task(task: TaskHeader, task_desc: str, raw_data: str):
        logger.info("translate_doc:llm_task:{}".format(task_desc))
        _submanager = SubManager(user_id=f"{task.sender.platform}:{task.sender.user_id}")
        driver = _submanager.llm_driver  # 由发送人承担接受者的成本
        model_name = os.getenv("OPENAI_API_MODEL", "gpt-3.5-turbo-0613")
        endpoint = openai.Openai(
            config=driver,
            model=model_name,
            temperature=0.1,
            messages=Message.create_short_task(
                task_desc=task_desc,
                refer=raw_data,
            ),
        )
        # 调用Openai
        result = await endpoint.create()
        await _submanager.add_cost(
            cost=UserInfo.Cost(token_usage=result.usage.total_tokens, token_uuid=driver.uuid, model_name=model_name)
        )
        assert result.default_message.content, "llm_task.py:llm_task:content is None"
        return result.default_message.content

    async def translate_docs(self, task: TaskHeader, file: File.Data, target_lang: str):
        if not file.file_name.endswith(('md', "txt")):
            raise ValueError("That Type File is Not supported :-(")

        from unstructured.partition.auto import partition
        elements = partition(file=BytesIO(initial_bytes=file.file_data), include_page_breaks=True)
        write_out_name = f"translated_{file.file_name}"

        write_out_file = BytesIO()
        write_out_file.name = write_out_name
        write_out_list = []

        async def _fill_box(text):
            try:
                await asyncio.sleep(2)
                result = await self.llm_task(task=task, task_desc=f"Translate text to {target_lang},as origin format",
                                             raw_data=text)
            except Exception as e:
                logger.error(e)
                result = str(element)
            write_out_list.append(result)

        _buffer = []
        for element in elements:
            _buffer.append(str(element))
            if len("/n".join(_buffer)) >= 1000:
                await _fill_box(text="\n".join(_buffer))
                _buffer = []
        if _buffer:
            await _fill_box(text="\n".join(_buffer))
            _buffer = []

        write_out_file.write("\n\n".join(write_out_list).encode("utf-8"))
        write_out_file.seek(0)
        return write_out_file

    async def callback(self, sign: str, task: TaskHeader):
        if sign == "reply":
            chain: Chain = await CHAIN_MANAGER.get_task(user_id=str(task.receiver.user_id))
            if chain:
                logger.info(f"{__plugin_name__}:chain callback locate in {sign} be sent")
                await Task(queue=chain.address).send_task(task=chain.arg)
            return True
        else:
            return False

    async def run(self, task: TaskHeader, receiver: TaskHeader.Location, arg, **kwargs):
        """
        处理message，返回message
        """
        try:
            _translate_file = []
            for item in task.message:
                if item.file:
                    for i in item.file:
                        _translate_file.append(i)
            try:
                translate_arg = Translate.parse_obj(arg)
            except Exception:
                raise ValueError("Please specify the following parameters clearly\n file_id=xxx,language=xxx")
            _file_obj = [await RawMessage.download_file(file_id=i.file_id)
                         for i in sorted(set(_translate_file), key=_translate_file.index)]
            _file_obj: List[File.Data] = [item for item in _file_obj if item]

            # 处理文件
            _result: List[File] = []
            if not _file_obj:
                return None
            for item in _file_obj:
                translated_file = await self.translate_docs(task=task, file=item, target_lang=translate_arg.language)
                file_obj = await RawMessage.upload_file(name=translated_file.name, data=translated_file.getvalue())
                _result.append(file_obj)
            # META
            _meta = task.task_meta.child(__plugin_name__)
            _meta.callback_forward = True
            _meta.callback_forward_reprocess = False
            _meta.callback = TaskHeader.Meta.Callback(
                role="function",
                name=__plugin_name__
            )
            await Task(queue=receiver.platform).send_task(
                task=TaskHeader(
                    sender=task.sender,
                    receiver=receiver,
                    task_meta=_meta,
                    message=[
                        RawMessage(
                            user_id=receiver.user_id,
                            chat_id=receiver.chat_id,
                            file=_result,
                            text="🍖 操作成功！"
                        )
                    ]
                )
            )

            logger.debug("Plugin:translate_doc say: {}".format(translate_arg))
        except Exception as e:
            logger.exception(e)
            await self.failed(platform=receiver.platform, task=task, receiver=receiver, reason=str(e))


__plugin_meta__ = PluginMetadata(
    name=__plugin_name__,
    description="Translate readable file to target language",
    usage=str(TranslateTool().keywords),
    openapi_version="20231013",
    function={
        FuncPair(function=translate, tool=TranslateTool)
    },
)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/17 下午8:46
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
import os
import ssl
import time
from typing import List

import telebot
from aio_pika.abc import AbstractIncomingMessage
from llmkira.middleware.chain_box import Chain, CHAIN_MANAGER
from llmkira.middleware.llm_task import OpenaiMiddleware
from llmkira.receiver import function
from llmkira.schema import RawMessage
from llmkira.sdk.endpoint import openai
from llmkira.sdk.error import RateLimitError
from llmkira.sdk.func_calling.register import ToolRegister
from llmkira.sdk.schema import Message, File
from llmkira.setting.telegram import BotSetting
from llmkira.task import Task, TaskHeader
from llmkira.utils import sync
from loguru import logger
from telebot import TeleBot, formatting

__receiver__ = "telegram"

from llmkira.middleware.router.schema import router_set

router_set(role="receiver", name=__receiver__)


class TelegramSender(object):
    """
    平台路由
    """

    def __init__(self):
        self.bot = TeleBot(token=BotSetting.token)
        from telebot import apihelper
        if BotSetting.proxy_address:
            apihelper.proxy = {'https': BotSetting.proxy_address}
        else:
            apihelper.proxy = None

    def forward(self, chat_id, reply_to_message_id, message: List[RawMessage]):
        """
        直接转发结果出来，不做任何处理
        """
        for item in message:
            for file_obj in item.file:
                if file_obj.file_url:
                    self.bot.send_document(chat_id=chat_id, document=file_obj.file_url,
                                           reply_to_message_id=reply_to_message_id, caption=file_obj.file_name)
                    continue
                _data: File.Data = sync(RawMessage.download_file(file_obj.file_id))

                if not _data:
                    logger.error(f"file download failed {file_obj.file_id}")
                    continue
                if file_obj.file_name.endswith(".jpg") or file_obj.file_name.endswith(".png"):
                    self.bot.send_photo(
                        chat_id=chat_id,
                        photo=_data.pair,
                        reply_to_message_id=reply_to_message_id,
                        caption=file_obj.file_name
                    )
                elif file_obj.file_name.endswith(".ogg"):
                    self.bot.send_voice(
                        chat_id=chat_id,
                        voice=_data.pair,
                        reply_to_message_id=reply_to_message_id,
                        caption=file_obj.file_name
                    )
                else:
                    self.bot.send_document(
                        chat_id=chat_id,
                        document=_data.pair, reply_to_message_id=reply_to_message_id,
                        caption=file_obj.file_name
                    )
            try:
                self.bot.send_message(
                    chat_id=chat_id,
                    text=item.text,
                    reply_to_message_id=reply_to_message_id,
                    parse_mode="MarkdownV2"
                )
            except telebot.apihelper.ApiTelegramException as e:
                time.sleep(3)
                logger.error("telegram send message error, retry")
                self.bot.send_message(
                    chat_id=chat_id,
                    text=item.text,
                    reply_to_message_id=reply_to_message_id
                )

    def reply(self, chat_id, reply_to_message_id, message: List[Message]):
        for item in message:
            assert item.content, f"message content is empty"
            self.bot.send_message(
                chat_id=chat_id,
                text=item.content,
                reply_to_message_id=reply_to_message_id
            )

    def error(self, chat_id, reply_to_message_id, text):
        self.bot.send_message(
            chat_id=chat_id,
            text=text,
            reply_to_message_id=reply_to_message_id
        )

    async def function(self, chat_id, reply_to_message_id,
                       task: TaskHeader,
                       llm: OpenaiMiddleware,
                       result: openai.OpenaiResult,
                       message: Message):
        if not message.function_call:
            raise ValueError("message not have function_call,forward type error")

        # 获取设置查看是否静音
        _tool = ToolRegister().get_tool(message.function_call.name)
        if not _tool:
            logger.warning(f"not found function {message.function_call.name}")
            return None
        task_message = formatting.format_text(
            formatting.mbold("🦴 Task be created:") + f" `{message.function_call.name}` ",
            formatting.mcode(message.function_call.arguments),
            separator="\n"
        )
        if not _tool().silent:
            self.bot.send_message(
                chat_id=chat_id,
                text=task_message,
                reply_to_message_id=reply_to_message_id,
                parse_mode="MarkdownV2"
            )
        # 回写创建消息
        sign = f"<{task.task_meta.sign_as[0] + 1}>"
        # 二周目消息不回写，因为写过了
        llm.write_back(
            role="assistant",
            name=message.function_call.name,
            message_list=[
                RawMessage(
                    text=f"Okay,Task be created:{message.function_call.arguments}.")]
        )
        # 构建对应的消息
        receiver = task.receiver.copy()
        receiver.platform = __receiver__

        # 运行函数
        await Task(queue=function.__receiver__).send_task(
            task=TaskHeader.from_function(
                parent_call=result,
                task_meta=task.task_meta,
                receiver=receiver,
                message=task.message
            )
        )


__sender__ = TelegramSender()


class TelegramReceiver(object):
    """
    receive message from telegram
    """

    def __init__(self):
        self.task = Task(queue=__receiver__)

    @staticmethod
    async def llm_request(llm_agent: OpenaiMiddleware, disable_function: bool = False):
        """
        校验包装，没有其他作用
        """
        try:
            _result = await llm_agent.request_openai(disable_function=disable_function)
            _message = _result.default_message
            logger.debug(f"[x] LLM Message Sent \n--message {_message}")
            assert _message, "message is empty"
            return _result
        except ssl.SSLSyscallError as e:
            logger.error(f"Network ssl error: {e},that maybe caused by bad proxy")
            raise e
        except RateLimitError as e:
            logger.error(f"ApiEndPoint:{e}")
            raise ValueError(f"Authentication expiration, overload or other issues with the Api Endpoint")
        except Exception as e:
            logger.exception(e)
            raise e

    async def _flash(self,
                     task: TaskHeader,
                     llm: OpenaiMiddleware,
                     auto_write_back: bool = True,
                     intercept_function: bool = False,
                     disable_function: bool = False
                     ):
        """
        函数池刷新
        :param auto_write_back: 是否将task携带的消息回写进消息池中，如果为False则丢弃task携带消息
        :param intercept_function: 是否拦截函数调用转发到函数处理器
        """
        try:
            llm.build(auto_write_back=auto_write_back)
            try:
                result = await self.llm_request(llm, disable_function=disable_function)
            except Exception as e:
                __sender__.error(
                    chat_id=task.receiver.chat_id,
                    reply_to_message_id=task.receiver.message_id,
                    text=f"🦴 Sorry, your request failed because: {e}"
                )
                return
            if intercept_function:
                # 拦截函数调用
                if hasattr(result.default_message, "function_call"):
                    return await __sender__.function(
                        chat_id=task.receiver.chat_id,
                        reply_to_message_id=task.receiver.message_id,
                        task=task,
                        llm=llm,  # IMPORTANT
                        message=result.default_message,
                        result=result
                    )
            return __sender__.reply(
                chat_id=task.receiver.chat_id,
                reply_to_message_id=task.receiver.message_id,
                message=[result.default_message]
            )
        except Exception as e:
            raise e

    async def deal_message(self, message):
        """
        处理消息
        """
        # 解析数据
        _task: TaskHeader = TaskHeader.parse_raw(message.body)
        # 没有任何参数
        if _task.task_meta.direct_reply:
            __sender__.forward(
                chat_id=_task.receiver.chat_id,
                reply_to_message_id=_task.receiver.message_id,
                message=_task.message
            )
            return None, None, None
            # 函数重整策略
        functions = None
        if _task.task_meta.function_enable:
            # 继承函数
            functions = _task.task_meta.function_list
            if _task.task_meta.sign_as[0] == 0:
                # 复制救赎
                _task.task_meta.function_salvation_list = _task.task_meta.function_list
                functions = []

                # 重整
                for _index, _message in enumerate(_task.message):
                    functions.extend(
                        ToolRegister().filter_pair(
                            key_phrases=_message.text,
                        )
                    )
                _task.task_meta.function_list = functions
        if _task.task_meta.sign_as[0] == 0:
            # 容错一层旧节点
            functions.extend(_task.task_meta.function_salvation_list)
        # 构建通信代理
        _llm = OpenaiMiddleware(task=_task, function=functions)
        logger.debug(f"[x] Received Order \n--order {_task.json(indent=2)}")
        # 插件直接转发与重处理
        if _task.task_meta.callback_forward:
            # 手动追加插件产生的线索消息
            _llm.write_back(
                role=_task.task_meta.callback.role,
                name=_task.task_meta.callback.name,
                message_list=_task.message
            )
            # 插件数据响应到前端
            if _task.task_meta.callback_forward_reprocess:
                # 手动写回则禁用从 Task 数据体自动回写
                # 防止AI去启动其他函数，禁用函数
                await self._flash(
                    llm=_llm,
                    task=_task,
                    intercept_function=True,
                    disable_function=True,
                    auto_write_back=False
                )
                # 同时递交部署点
                return _task, _llm, "callback_forward_reprocess"
            # 转发函数
            __sender__.forward(
                chat_id=_task.receiver.chat_id,
                reply_to_message_id=_task.receiver.message_id,
                message=_task.message
            )
            # 同时递交部署点
            return _task, _llm, "callback_forward_reprocess"
        await self._flash(llm=_llm, task=_task, intercept_function=True)
        return None, None, None

    async def on_message(self, message: AbstractIncomingMessage):
        # 过期时间
        # print(message.expiration)
        try:
            if os.getenv("LLMBOT_STOP_REPLY") == "1":
                return None
            _task, _llm, _point = await self.deal_message(message)
            # 启动链式函数应答循环
            if _task:
                chain: Chain = await CHAIN_MANAGER.get_task(user_id=str(_task.receiver.user_id))
                if chain:
                    logger.info(f"Catch chain callback\n--callback_send_by {_point}")
                    await Task(queue=chain.address).send_task(task=chain.arg)
        except Exception as e:
            logger.exception(e)
            await message.reject(requeue=False)
        finally:
            await message.ack(multiple=False)

    async def telegram(self):
        if not BotSetting.available:
            logger.warning("Receiver Runtime:TelegramBot Setting not available")
            return None
        await self.task.consuming_task(self.on_message)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/17 下午8:18
# @Author  : sudoskys
# @File    : app.py
# @Software: PyCharm
import asyncio
import os
import sys

from dotenv import load_dotenv
from llmkira import load_plugins
from llmkira.sdk.func_calling import load_from_entrypoint, get_entrypoint_plugins
from loguru import logger

from .rss import RssApp
from .telegram import TelegramBotRunner

load_dotenv()
logger.remove()
handler_id = logger.add(sys.stderr,
                        level="INFO" if os.getenv("LLMBOT_LOG_OUTPUT") != "DEBUG" else "DEBUG",
                        enqueue=True
                        )
logger.add(sink='run.log',
           format="{time} - {level} - {message}",
           level="INFO",
           rotation="100 MB",
           enqueue=True
           )

__area__ = "sender"

# 注册机器人事件
telegram_bot = TelegramBotRunner().telegram()
rss_app = RssApp()

func = [
    telegram_bot,
    rss_app.rss_polling(interval=60 * 60 * 1),
]

# 初始化插件系统
load_plugins("llmkira/extra/plugins")
load_from_entrypoint("llmkira.extra.plugin")

loaded_message = "\n >>".join(get_entrypoint_plugins())
logger.success(f"\n===========Third Party Plugins Loaded==========\n >>{loaded_message}")


async def _main():
    await asyncio.gather(
        *func
    )


loop = asyncio.get_event_loop()
loop.run_until_complete(_main())



# -*- coding: utf-8 -*-
# @Time    : 2023/10/11 下午3:34
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
import json
from typing import Optional

from loguru import logger

from llmkira.cache.redis import cache
from llmkira.task import TaskHeader
from .schema import Chain


class AuthReloader(object):

    async def add_auth(self, task: Chain):
        _cache = await cache.set_data(key=f"auth:{task.uuid}", value=task.json(), timeout=60 * 60 * 24 * 7)
        return task.uuid

    async def get_auth(self, uuid: str, user_id: str) -> Optional[Chain]:
        _cache = await cache.read_data(key=f"auth:{uuid}")
        if not _cache:
            logger.debug(f"[x] Auth \n--empty {uuid}")
            return None
        logger.debug(f"[x] Auth \n--data {_cache}")
        task = Chain().parse_obj(_cache)
        task = task.format_arg(arg=TaskHeader)
        if task.user_id != user_id:
            logger.debug(f"[x] Auth \n--not found user {uuid}")
            return None
        return task


class ChainReloader(object):

    async def add_task(self, task: Chain):
        _cache = await cache.lpush_data(key=f"chain:{task.user_id}", value=task.json())
        return task.uuid

    async def get_task(self, user_id: str) -> Optional[Chain]:
        _data = await cache.lpop_data(key=f"chain:{user_id}")
        if not _data:
            return None
        _task = Chain.parse_obj(json.loads(_data))
        _task = _task.format_arg(arg=TaskHeader)
        return _task


AUTH_MANAGER = AuthReloader()
CHAIN_MANAGER = ChainReloader()



# -*- coding: utf-8 -*-
# @Time    : 2023/8/16 上午11:06
# @Author  : sudoskys
# @File    : components.py
# @Software: PyCharm
import asyncio
# ATTENTION:禁止调用上层任何包，否则会导致循环引用
from typing import Literal, Optional, Coroutine, List

from loguru import logger
from pydantic import BaseModel, root_validator, Field

from llmkira.sdk.error import ValidationError


def _sync(coroutine: Coroutine):
    """
    https://nemo2011.github.io/bilibili-api/#/sync-executor
    """
    try:
        asyncio.get_event_loop()
    except Exception as e:
        asyncio.set_event_loop(asyncio.new_event_loop())
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(coroutine)


class Function(BaseModel):
    """
    函数定义体。
    供外部模组定义并注册函数
    """

    class Parameters(BaseModel):
        type: str = "object"
        properties: dict = {}

    name: str = Field(None, description="函数名称", regex=r"^[a-zA-Z0-9_]+$")
    description: Optional[str] = None
    parameters: Parameters = Parameters(type="object")
    required: List[str] = []

    def add_property(self, property_name: str,
                     property_type: Literal["string", "integer", "number", "boolean", "object"],
                     property_description: str,
                     enum: Optional[tuple] = None,
                     required: bool = False
                     ):
        self.parameters.properties[property_name] = {}
        self.parameters.properties[property_name]['type'] = property_type
        self.parameters.properties[property_name]['description'] = property_description
        if enum:
            self.parameters.properties[property_name]['enum'] = tuple(enum)
        if required:
            self.required.append(property_name)

    def parse_schema_to_properties(self, schema: BaseModel):
        """
        解析pydantic的schema
        传入一个pydantic的schema，解析成properties
        参数可以被数据模型所定义
        """
        self.parameters.properties = schema.schema()["properties"]


class File(BaseModel):
    class Data(BaseModel):
        file_name: str
        file_data: bytes = Field(None, description="文件数据")

        @property
        def pair(self):
            return self.file_name, self.file_data

    file_id: str = Field(None, description="文件ID")
    file_name: str = Field(None, description="文件名")
    file_url: str = Field(None, description="文件URL")

    # hash able
    def __eq__(self, other):
        if isinstance(other, File):
            return (self.file_id == other.file_id) and (self.file_name == other.file_name)
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.file_id) + hash(self.file_name)

    @property
    def file_prompt(self):
        return f"[ReadableFile(file_name={self.file_name}, file_id={self.file_id})]"


class Message(BaseModel):
    """
    标准 API 件
    """

    class FunctionCall(BaseModel):
        name: str
        arguments: str

    role: Literal["system", "assistant", "user", "function"] = "user"
    content: Optional[str] = None
    # speaker
    name: Optional[str] = Field(None, description="speaker_name", regex=r"^[a-zA-Z0-9_]+$")
    # AI generated function call
    function_call: Optional[FunctionCall] = None

    @root_validator
    def check(cls, values):
        if values.get("role") == "function" and not values.get("name"):
            raise ValidationError("sdk param validator:name must be specified when role is function")
        if not values.get("content") and not values.get("function_call"):
            raise ValidationError("sdk param validator:content or function_call must be specified")
        # 过滤value中的None
        return {k: v for k, v in values.items() if v is not None}

    @property
    def message(self):
        return self

    @classmethod
    def create_short_task(cls, task_desc, refer, role: str = None):
        """
        生成task
        """
        if not role:
            role = "[RULE]Please complete the order according to the task description refer to given information, if can't complete, please reply 'give up'"
        return [
            Message(
                role="system",
                content=role,
            ),
            Message(
                role="user",
                content=f"{refer} <hint>{task_desc}<hint>",
                name="task"
            )
        ]

    @classmethod
    def create_task_message_list(cls, task_desc, refer, role: str = None):
        """
        生成task
        """
        if not role:
            role = "[Scene: chatting in real time]\nPlease complete the order according to the task description refer to given information"
        return [
            Message(
                role="system",
                content=role,
            ),
            Message(
                role="assistant",
                content=refer,
                name="information"
            ),
            Message(
                role="user",
                content=task_desc,
                name="task"
            ),
            Message(
                role="assistant",
                content="ok , i will complete the tasks you give",
            )
        ]


class TextMessage(Message):
    """
    存储类型：普通文本消息
    """

    class FunctionCall(BaseModel):
        name: str
        arguments: str

    message_type = "message"

    ##

    role: Literal["system", "assistant", "user", "function"] = "user"
    content: Optional[str] = None
    # speaker
    name: Optional[str] = Field(None, description="speaker_name", regex=r"^[a-zA-Z0-9_]+$")
    # AI generated function call
    function_call: Optional[FunctionCall] = None

    @root_validator
    def check(cls, values):
        if values.get("role") == "function" and not values.get("name"):
            raise ValidationError("sdk param validator:name must be specified when role is function")
        if not values.get("content") and not values.get("function_call"):
            raise ValidationError("sdk param validator:content or function_call must be specified")
        # 过滤value中的None
        return {k: v for k, v in values.items() if v is not None}

    @property
    def message(self) -> Message:
        return Message(
            role=self.role,
            name=self.name,
            content=self.content,
            function_call=self.function_call
        )


class FileMessage(Message):
    """
    多媒体消息

    :param role: 角色
    :param metadata: 元数据

    """
    message_type = "file_message"
    # 参数固定
    role: Literal["system", "assistant", "user", "function"] = "user"
    name: Optional[str] = Field("action", description="speaker_name", regex=r"^[a-zA-Z0-9_]+$")
    content: str = Field(None, description="内容")

    file: File = Field(None, description="文件")
    caption: str = Field("", description="注释")
    metadata: Optional[dict] = None

    @root_validator
    def check(cls, values):
        if not values.get("file"):
            raise ValidationError("sdk param validator:file must be specified")
        return values

    @property
    def message(self) -> Message:
        assert hasattr(self.file, "file_name"), "obj not ok"
        content = f"(Operable File)[{self.file.file_name}]/{self.caption[:20]}.../"
        # 最终向API发送的消息
        return Message(
            role=self.role,
            name=self.name,
            content=content,
            function_call=None
        )


def standardise(message: Message):
    if not isinstance(message, Message):
        raise TypeError(f"message must be Message, not {type(message)}")
    if hasattr(message, "message"):
        return message.message
    else:
        return message


def parse_message_dict(item: dict):
    message_type = item.get("message_type", None)
    if message_type == "message":
        _message = TextMessage.parse_obj(item)
    elif message_type == "file_message":
        _message = FileMessage.parse_obj(item)
    else:
        try:
            _message = Message.parse_obj(item)
        except Exception as e:
            logger.error(f"parse_message_dict:Unknown message type in redis data {e}")
            # raise ValueError(f"Unknown message type in redis data {message_type}")
        else:
            return _message



# -*- coding: utf-8 -*-
# 探针插件加载机制 参考 nonebot2 项目设计
# https://github.com/nonebot/nonebot2/blob/99931f785a31138a2f6bac1d103551dab47d40f7/nonebot/plugin/manager.py

from contextvars import ContextVar
from itertools import chain
from pathlib import Path
from types import ModuleType
from typing import Optional, List, Set, Dict, Tuple, Any

from pydantic import BaseModel

from .error import OpenApiError

_plugins: Dict[str, "Plugin"] = {}
_managers: List["PluginManager"] = []
_current_plugin_chain: ContextVar[Tuple["Plugin", ...]] = ContextVar(
    "_current_plugin_chain", default=tuple()
)


class FrameworkInfo(BaseModel):
    support: bool
    exception: Optional[str] = None


_current_openapi_version_: str = "20231013"
_openapi_version_: Dict[str, "FrameworkInfo"] = {
    _current_openapi_version_: FrameworkInfo(
        support=True,
        exception=None
    ),
}


def verify_openapi_version(name: str, metadata: "PluginMetadata"):
    """
    验证框架接口版本
    """
    if not metadata:
        return None
    frame = _openapi_version_.get(metadata.openapi_version, None)
    if not frame:
        raise OpenApiError(
            f"Plugin:{name}\n--error {metadata.openapi_version} not support")
    if not frame.support:
        raise OpenApiError(
            f"Plugin:{name}\n--error {frame.exception}")


def path_to_module_name(path: Path) -> str:
    rel_path = path.resolve().relative_to(Path("").resolve())
    if rel_path.stem == "__init__":
        return ".".join(rel_path.parts[:-1])
    else:
        return ".".join(rel_path.parts[:-1] + (rel_path.stem,))


def get_plugin(name: str) -> Optional["Plugin"]:
    """获取已经导入的某个插件。

    如果为 `load_plugins` 文件夹导入的插件，则为文件(夹)名。

    参数:
        name: 插件名，即 {ref}`extra.plugin.model.Plugin.name`。
    """
    return _plugins.get(name)


def get_loaded_plugins() -> Set["Plugin"]:
    """获取当前已导入的所有插件。"""
    return set(_plugins.values())


def get_entrypoint_plugins(group="llmkira.extra.plugin") -> Set[str]:
    import importlib_metadata
    hook = importlib_metadata.entry_points().select(group=group)
    plugins = [item.module for item in hook]
    return {*plugins}


def get_available_plugin_names() -> Set[str]:
    """获取当前所有可用的插件名（包含尚未加载的插件）。"""
    return {*chain.from_iterable(manager.available_plugins for manager in _managers)}


def _find_manager_by_name(name: str) -> Optional[Any]:
    for manager in reversed(_managers):
        if name in manager.plugins or name in manager.searched_plugins:
            return manager


def _module_name_to_plugin_name(module_name: str) -> str:
    return module_name.rsplit(".", 1)[-1]


def _new_plugin(module_name: str, module: ModuleType, manager: "PluginManager") -> "Plugin":
    """
    Create a new plugin
    """
    plugin_name = _module_name_to_plugin_name(module_name)
    if plugin_name in _plugins:
        raise RuntimeError("Plugin already exists! Check your plugin name.")
    plugin = Plugin(plugin_name, module, module_name, manager)
    _plugins[plugin_name] = plugin
    return plugin


def _revert_plugin(plugin: "Plugin") -> None:
    """
    Revert a plugin
    删除创建的链
    :param plugin: Plugin to revert
    """
    if plugin.name not in _plugins:
        raise RuntimeError("Plugin not found!")
    del _plugins[plugin.name]
    parent_plugin = plugin.parent_plugin
    if parent_plugin:
        parent_plugin.sub_plugins.remove(plugin)


from .loader import PluginManager
from .schema import Plugin as Plugin, BaseTool
from .loader import require as require
from .loader import load_plugin as load_plugin
from .loader import load_plugins as load_plugins
from .loader import load_all_plugins as load_all_plugins
from .loader import load_builtin_plugin as load_builtin_plugin
from .loader import load_builtin_plugins as load_builtin_plugins
from .loader import load_from_entrypoint as load_from_entrypoint
from .loader import PluginManager as PluginManager

from .model import Plugin as Plugin
from .model import PluginMetadata as PluginMetadata

from .register import ToolRegister as ToolRegister



# -*- coding: utf-8 -*-
# @Time    : 2023/8/18 下午6:11
# @Author  : sudoskys
# @Software: PyCharm
import datetime
import re
import time

from loguru import logger
from pydantic import validator, BaseModel

from llmkira.middleware.chain_box import Chain, CHAIN_MANAGER
from llmkira.receiver.aps import SCHEDULER
from llmkira.schema import RawMessage
from llmkira.sdk.endpoint.openai import Function
from llmkira.sdk.func_calling import PluginMetadata, BaseTool, verify_openapi_version
from llmkira.sdk.func_calling.schema import FuncPair
from llmkira.task import Task, TaskHeader

__plugin_name__ = "set_alarm_reminder"
alarm = Function(name=__plugin_name__, description="Set a timed reminder (only for minutes)")
alarm.add_property(
    property_name="delay",
    property_description="The delay time, in minutes",
    property_type="integer",
    required=True
)
alarm.add_property(
    property_name="content",
    property_description="reminder content",
    property_type="string",
    required=True
)


class Alarm(BaseModel):
    delay: int
    content: str

    class Config:
        extra = "allow"

    @validator("delay")
    def delay_validator(cls, v):
        if v < 0:
            raise ValueError("delay must be greater than 0")
        return v


class AlarmTool(BaseTool):
    """
    搜索工具
    """
    silent: bool = False
    function: Function = alarm
    keywords: list = ["闹钟", "提醒", "定时", "到点", '分钟']
    pattern = re.compile(r"(\d+)(分钟|小时|天|周|月|年)后提醒我(.*)")
    require_auth: bool = True

    def pre_check(self):
        return True

    def func_message(self, message_text):
        """
        如果合格则返回message，否则返回None，表示不处理
        """
        for i in self.keywords:
            if i in message_text:
                return self.function
        # 正则匹配
        if self.pattern:
            match = self.pattern.match(message_text)
            if match:
                return self.function
        return None

    async def failed(self, platform, task, receiver, reason):
        try:

            _meta = task.task_meta.child(__plugin_name__)
            _meta.callback_forward = True
            _meta.reprocess_needed = False
            _meta.callback = TaskHeader.Meta.Callback(
                role="function",
                name=__plugin_name__
            )

            await Task(queue=platform).send_task(
                task=TaskHeader(
                    sender=task.sender,
                    receiver=receiver,
                    task_meta=_meta,
                    message=[
                        RawMessage(
                            user_id=receiver.user_id,
                            chat_id=receiver.chat_id,
                            text=f"🍖 {__plugin_name__}操作失败了！原因：{reason}"
                        )
                    ]
                )
            )
        except Exception as e:
            logger.error(e)

    async def callback(self, sign: str, task: TaskHeader):
        if sign == "reply":
            chain: Chain = await CHAIN_MANAGER.get_task(user_id=str(task.receiver.user_id))
            if chain:
                logger.info(f"{__plugin_name__}:chain callback locate in {sign} be sent")
                await Task(queue=chain.address).send_task(task=chain.arg)
            return True
        else:
            return False

    async def run(self, task: TaskHeader, receiver: TaskHeader.Location, arg, **kwargs):
        """
        处理message，返回message
        """
        try:
            _set = Alarm.parse_obj(arg)
            #
            _meta = task.task_meta.child(__plugin_name__)
            _meta.callback_forward = True
            _meta.callback_forward_reprocess = False
            _meta.callback = TaskHeader.Meta.Callback(
                role="function",
                name=__plugin_name__
            )

            async def _send(receiver, _set):
                await Task(queue=receiver.platform).send_task(
                    task=TaskHeader(
                        sender=task.sender,  # 继承发送者
                        receiver=receiver,  # 因为可能有转发，所以可以单配
                        task_meta=_meta,
                        message=[
                            RawMessage(
                                user_id=receiver.user_id,
                                chat_id=receiver.chat_id,
                                text=_set.content
                            )
                        ]
                    )
                )

            logger.debug("Plugin:set alarm {} minutes later".format(_set.delay))
            SCHEDULER.add_job(
                func=_send,
                id=str(time.time()),
                trigger="date",
                replace_existing=True,
                run_date=datetime.datetime.now() + datetime.timedelta(minutes=_set.delay),
                args=[receiver, _set]
            )
            try:
                SCHEDULER.start()
            except Exception as e:
                pass
            await Task(queue=receiver.platform).send_task(
                task=TaskHeader(
                    sender=task.sender,  # 继承发送者
                    receiver=receiver,  # 因为可能有转发，所以可以单配
                    task_meta=_meta,
                    message=[
                        RawMessage(
                            user_id=receiver.user_id,
                            chat_id=receiver.chat_id,
                            text=f"🍖 The alarm is now set,just wait for {_set.delay} min!"
                        )
                    ]
                )
            )
        except Exception as e:
            logger.exception(e)
            await self.failed(platform=receiver.platform, task=task, receiver=receiver, reason=str(e))


__plugin_meta__ = PluginMetadata(
    name=__plugin_name__,
    description="Set a timed reminder (only for minutes)",
    usage="set_alarm_reminder 10 minutes later remind me to do something",
    openapi_version="20231013",
    function={
        FuncPair(function=alarm, tool=AlarmTool)
    },
    homepage="https://github.com/LlmKira"
)
verify_openapi_version(__plugin_name__, __plugin_meta__)



# -*- coding: utf-8 -*-
from typing import Tuple, Union

import cjieba
from sklearn.metrics.pairwise import cosine_similarity

from ..solo import singleton
from ...api.keywords import STOPWORDS


@singleton
class CosionSimilarity(object):
    """
    根据余弦函数计算相似性
    one-hot编码
    """

    def __init__(self):
        self.stopwords = self.load_stopwords(STOPWORDS)

    @staticmethod
    def load_stopwords(stopwords_path):
        with open(stopwords_path, 'r', encoding='utf-8') as f:
            return set([line.strip() for line in f])

    def cut_words(self, text):
        return [word for word in cjieba.cut(text) if word not in self.stopwords]

    def str_to_vector(self, text1: str, text2: str) -> Tuple[list, list]:
        text1_words = set(self.cut_words(text1))
        text2_words = set(self.cut_words(text2))
        all_words = list(text1_words | text2_words)
        text1_vector = [1 if word in text1_words else 0 for word in all_words]
        text2_vector = [1 if word in text2_words else 0 for word in all_words]
        return text1_vector, text2_vector

    def similarity(self, text1: Union[str, list], text2: Union[str, list]):
        text1_words = set(self.cut_words(text1))
        text2_words = set(self.cut_words(text2))
        all_words = list(text1_words | text2_words)
        text1_vector = [1 if word in text1_words else 0 for word in all_words]
        text2_vector = [1 if word in text2_words else 0 for word in all_words]
        if not text1_vector or not text2_vector:
            return 0
        return cosine_similarity([text1_vector], [text2_vector])[0][0]

    @staticmethod
    def vector_similarity(text1_vector: list, text2_vector: list):
        from sklearn.metrics.pairwise import cosine_similarity
        return cosine_similarity([text1_vector], [text2_vector])[0][0]


if __name__ == '__main__':
    text1 = "小明，你妈妈喊你回家吃饭啦"
    text2 = "回家吃饭啦，小明"
    similarity = CosionSimilarity()
    print(similarity.similarity(text1, text2))



# -*- coding: utf-8 -*-
# @Time    : 2023/7/10 下午9:44
# @Author  : sudoskys
# @File    : elara.py
# @Software: PyCharm
import asyncio

import elara

from llmkira.cache.base import AbstractDataClass, PREFIX


class ElaraClientWrapper(AbstractDataClass):
    """
    Elara 数据基类
    """

    def __init__(self, backend, prefix=PREFIX):
        self.prefix = prefix
        self.elara = elara.exe(path=backend)
        self.lock = asyncio.Lock()

    async def ping(self):
        return True

    def update_backend(self, backend):
        self.elara = elara.exe(path=backend)
        return True

    async def read_data(self, key):
        async with self.lock:
            _res = self.elara.get(self.prefix + str(key))
            return _res

    async def set_data(self, key, value, timeout=None):
        async with self.lock:
            self.elara.set(self.prefix + str(key), value)
            self.elara.commit()





# -*- coding: utf-8 -*-

"""
bilibili_api.utils.sync

同步执行异步函数
"""
import asyncio
import hashlib
from typing import Coroutine

import nest_asyncio

nest_asyncio.apply()


def __ensure_event_loop():
    try:
        asyncio.get_event_loop()
    except Exception as e:
        asyncio.set_event_loop(asyncio.new_event_loop())


def sync(coroutine: Coroutine):
    """
    同步执行异步函数，使用可参考 [同步执行异步代码](https://nemo2011.github.io/bilibili-api/#/sync-executor)

    Args:
        coroutine (Coroutine): 异步函数

    Returns:
        该异步函数的返回值
    """
    __ensure_event_loop()
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(coroutine)


def sha1_encrypt(string):
    """
    sha1加密算法
    """

    sha = hashlib.sha1(string.encode('utf-8'))
    encrypts = sha.hexdigest()
    return encrypts[:8]



# -*- coding: utf-8 -*-
# @Time    : 2023/8/15 下午10:29
# @Author  : sudoskys
# @File    : error.py
# @Software: PyCharm


class ValidationError(Exception):
    pass


class AuthenticationError(Exception):
    """
    Raised when the API key is invalid.
    """
    pass


class RateLimitError(Exception):
    """
    Raised when the API key has exceeded its rate limit.
    """
    pass


class ServiceUnavailableError(Exception):
    """
    Raised when the API is unavailable.
    """
    pass



# -*- coding: utf-8 -*-
# @Time    : 2023/10/13 下午1:25
# 探针插件加载机制 参考 nonebot2 项目设计
# https://github.com/nonebot/nonebot2/blob/99931f785a31138a2f6bac1d103551dab47d40f7/nonebot/plugin/manager.py

import importlib
import pkgutil
import sys
from importlib.abc import MetaPathFinder
from importlib.machinery import PathFinder, SourceFileLoader
from itertools import chain
from pathlib import Path
from types import ModuleType
from typing import Optional, Sequence, List, Set, Dict, Iterable

from loguru import logger

from . import (
    _managers,
    _new_plugin,
    _revert_plugin,
    _current_plugin_chain,
    _module_name_to_plugin_name,
    path_to_module_name
)
from .error import escape_tag
from .schema import Plugin, PluginMetadata


class PluginManager:
    """插件管理器。

    参数:
        plugins: 独立插件模块名集合。
        search_path: 插件搜索路径（文件夹）。
    """

    def __init__(
            self,
            plugins: Optional[Iterable[str]] = None,
            search_path: Optional[Iterable[str]] = None,
    ):
        # simple plugin not in search path
        self.plugins: Set[str] = set(plugins or [])
        self.search_path: Set[str] = set(search_path or [])

        # cache plugins
        self._third_party_plugin_names: Dict[str, str] = {}
        self._searched_plugin_names: Dict[str, Path] = {}
        self.prepare_plugins()

    def __repr__(self) -> str:
        return f"PluginManager(plugins={self.plugins}, search_path={self.search_path})"

    @property
    def third_party_plugins(self) -> Set[str]:
        """返回所有独立插件名称。"""
        return set(self._third_party_plugin_names.keys())

    @property
    def searched_plugins(self) -> Set[str]:
        """返回已搜索到的插件名称。"""
        return set(self._searched_plugin_names.keys())

    @property
    def available_plugins(self) -> Set[str]:
        """返回当前插件管理器中可用的插件名称。"""
        return self.third_party_plugins | self.searched_plugins

    def _previous_plugins(self) -> Set[str]:
        _pre_managers: List[PluginManager]
        if self in _managers:
            _pre_managers = _managers[: _managers.index(self)]
        else:
            _pre_managers = _managers[:]

        return {
            *chain.from_iterable(manager.available_plugins for manager in _pre_managers)
        }

    def prepare_plugins(self) -> Set[str]:
        """搜索插件并缓存插件名称。"""
        # get all previous ready to load plugins
        previous_plugins = self._previous_plugins()
        searched_plugins: Dict[str, Path] = {}
        third_party_plugins: Dict[str, str] = {}

        # check third party plugins
        for plugin in self.plugins:
            name = _module_name_to_plugin_name(plugin)
            if name in third_party_plugins or name in previous_plugins:
                raise RuntimeError(
                    f"Plugin already exists: {name}! Check your plugin name"
                )
            third_party_plugins[name] = plugin

        self._third_party_plugin_names = third_party_plugins
        # check plugins in search path
        for module_info in pkgutil.iter_modules(self.search_path):
            # ignore if startswith "_"
            if module_info.name.startswith("_"):
                continue

            if (
                    module_info.name in searched_plugins
                    or module_info.name in previous_plugins
                    or module_info.name in third_party_plugins
            ):
                raise RuntimeError(
                    f"Plugin already exists: {module_info.name}! Check your plugin name"
                )
            #
            module_spec = module_info.module_finder.find_spec(module_info.name, None)
            if not module_spec:
                continue
            module_path = module_spec.origin
            if not module_path:
                continue
            searched_plugins[module_info.name] = Path(module_path).resolve()
        self._searched_plugin_names = searched_plugins
        return self.available_plugins

    def load_plugin(self, name: str) -> Optional[Plugin]:
        """加载指定插件。

        对于独立插件，可以使用完整插件模块名或者插件名称。

        参数:
            name: 插件名称。
        """

        try:
            if name in self.plugins:
                module = importlib.import_module(name)
            elif name in self._third_party_plugin_names:
                module = importlib.import_module(self._third_party_plugin_names[name])
            elif name in self._searched_plugin_names:
                module = importlib.import_module(
                    path_to_module_name(self._searched_plugin_names[name])
                )
            else:
                raise RuntimeError(f"Plugin not found: {name}! Check your plugin name")
            plugin = getattr(module, "__plugin__", None)
            if plugin is None or not isinstance(plugin, Plugin):
                raise RuntimeError(
                    f"Module {module.__name__} is not loaded as a plugin! "
                    "Make sure not to import it before loading."
                )
            logger.opt(colors=True).success(
                f'Succeeded to load plugin "<y>{escape_tag(plugin.name)}</y>"'
                + (
                    f' from "<m>{escape_tag(plugin.module_name)}</m>"'
                    if plugin.module_name != plugin.name
                    else ""
                )
            )
            return plugin
        except Exception as e:
            logger.opt(colors=True, exception=e).error(
                f'<r><bg #f5f2f2>Failed to import "{escape_tag(name)}"</bg #f5f2f2></r>'
            )

    def load_all_plugins(self) -> Set[Plugin]:
        """加载所有可用插件。"""

        return set(
            filter(None, (self.load_plugin(name) for name in self.available_plugins))
        )


class PluginFinder(MetaPathFinder):
    def find_spec(
            self,
            fullname: str,
            path: Optional[Sequence[str]],
            target: Optional[ModuleType] = None,
    ):
        if _managers:
            module_spec = PathFinder.find_spec(fullname, path, target)
            if not module_spec:
                return
            module_origin = module_spec.origin
            if not module_origin:
                return
            module_path = Path(module_origin).resolve()

            for manager in reversed(_managers):
                # use path instead of name in case of submodule name conflict
                if (
                        fullname in manager.plugins
                        or module_path in manager._searched_plugin_names.values()
                ):
                    # print("find_spec",manager, fullname, module_origin)
                    # print(f"module_origin:{module_origin}")
                    module_spec.loader = PluginLoader(manager, fullname, module_origin)
                    return module_spec
        return


class PluginLoader(SourceFileLoader):
    def __init__(self,
                 manager: PluginManager, fullname: str, path) -> None:
        self.manager = manager
        self.loaded = False
        super().__init__(fullname, path)

    def create_module(self, spec) -> Optional[ModuleType]:
        # check self.name in sys.modules to avoid import loop
        if self.name in sys.modules:
            self.loaded = True
            return sys.modules[self.name]
        # return None to use default module creation
        return super().create_module(spec)

    def exec_module(self, module: ModuleType) -> None:
        if self.loaded:
            return

        # create plugin before executing
        plugin = _new_plugin(self.name, module, self.manager)
        setattr(module, "__plugin__", plugin)

        # detect parent plugin before entering current plugin context
        parent_plugins = _current_plugin_chain.get()
        for pre_plugin in reversed(parent_plugins):
            # ensure parent plugin is declared before current plugin
            if _managers.index(pre_plugin.manager) < _managers.index(self.manager):
                plugin.parent_plugin = pre_plugin
                pre_plugin.sub_plugins.add(plugin)
                break

        # enter plugin context
        _plugin_token = _current_plugin_chain.set(parent_plugins + (plugin,))

        try:
            super().exec_module(module)
        except Exception:
            _revert_plugin(plugin)
            raise
        except SystemExit:
            _revert_plugin(plugin)
            raise RuntimeError(
                f"Plugin {self.name} exited with status 0. "
                "Plugins should not call sys.exit()!"
            )
        finally:
            # leave plugin context
            _current_plugin_chain.reset(_plugin_token)
        # get plugin metadata
        metadata: Optional[PluginMetadata] = getattr(module, "__plugin_meta__", None)
        # 校验
        if not metadata:
            raise RuntimeError(
                f"Plugin {self.name} does not provide metadata! "
                "Make sure to declare it before importing."
            )

        plugin.metadata = metadata
        return


sys.meta_path.insert(0, PluginFinder())



# -*- coding: utf-8 -*-
# @Time    : 2023/7/10 下午9:43
# @Author  : sudoskys
# @File    : redis.py
# @Software: PyCharm
import json
import os
from typing import Optional, Tuple, List

import redis
from dotenv import load_dotenv
from llmkira.cache.base import AbstractDataClass, PREFIX
from loguru import logger
from redis.asyncio.client import Redis
from redis.asyncio.connection import ConnectionPool


class RedisClientWrapper(AbstractDataClass):
    """
    Redis 数据类
    """

    def __init__(self, backend, prefix=PREFIX):
        self.prefix = prefix
        self.connection_pool = redis.asyncio.ConnectionPool.from_url(backend)
        self._redis = redis.asyncio.Redis(connection_pool=self.connection_pool)

    async def ping(self):
        return await self._redis.ping()

    def update_backend(self, backend):
        self.connection_pool = ConnectionPool.from_url(backend)
        self._redis = Redis(connection_pool=self.connection_pool)
        return True

    async def set_data(self, key, value, timeout=None):
        if isinstance(value, (dict, list)):
            value = json.dumps(value)
        return await self._redis.set(name=f"{self.prefix}{key}", value=value, ex=timeout)

    async def read_data(self, key):
        data = await self._redis.get(self.prefix + str(key))
        if data is not None:
            try:
                data = json.loads(data)
            except Exception as e:
                pass
        return data

    async def lpush_data(self, key, value):
        """
        从左侧插入数据
        :param key: str
        :param value: json
        """
        # 验证是否可以被json序列化
        return await self._redis.lpush(self.prefix + str(key), value)

    async def lpop_data(self, key) -> Optional[str]:
        _data = await self._redis.lpop(self.prefix + str(key))
        if _data:
            return _data.decode("utf-8")
        return None

    async def lrange_data(self, key, start_end: Tuple[int, int] = (0, -1)) -> List[str]:
        _items = await self._redis.lrange(self.prefix + str(key), start=start_end[0], end=start_end[1])
        items = [m.decode("utf-8") for m in _items[::-1]]
        return items


def check_redis_dsn(dsn):
    try:
        import redis
        r = redis.from_url(dsn)
        assert r.ping() is True
    except Exception as exp:
        logger.warning(f"Error connecting to Redis: {exp}")
        return False
    else:
        return True


# 加载 .env 文件
load_dotenv()
redis_url = os.getenv('REDIS_DSN', None)
if not redis_url:
    logger.warning('REDIS_DSN not found in .env, use default redis://localhost:6379/0')
    redis_url = 'redis://localhost:6379/0'
cache: Optional[RedisClientWrapper]
if not check_redis_dsn(redis_url):
    logger.warning('REDIS DISCONNECT,pls set REDIS_DSN in env')
    cache = None
else:
    logger.success(f'RedisClientWrapper loaded successfully in {redis_url}')
    cache = RedisClientWrapper(redis_url)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/16 下午8:54
# @Author  : sudoskys
# @File    : func_call.py
# @Software: PyCharm
import threading
from typing import List, Dict
from typing import Optional, Type

from . import _openapi_version_, BaseTool, get_loaded_plugins
from .schema import FuncPair, Function

threading_lock = threading.Lock()


class ToolRegister(object):
    """
    扩展对 _plugins 字段的操作,需要实例化以获取数据
    """

    def __init__(self):
        self.version = _openapi_version_
        self.pair_function: Dict[str, FuncPair] = {}
        self.prepare()

    def prepare(self):
        for item in get_loaded_plugins():
            for sub_item in item.metadata.function:
                self.pair_function[sub_item.name] = sub_item

    def get_tool(self, name: str) -> Optional[Type[BaseTool]]:
        if not self.pair_function.get(name, None):
            return None
        return self.pair_function[name].tool

    @property
    def functions(self) -> Dict[str, Function]:
        _item: Dict[str, Function] = {}
        for item in get_loaded_plugins():
            for sub_item in item.metadata.function:
                _item[sub_item.name] = sub_item.function
        return _item

    @property
    def tools(self) -> List[Type[BaseTool]]:
        _item: List[Type[BaseTool]] = []
        for item in get_loaded_plugins():
            for sub_item in item.metadata.function:
                _item.append(sub_item.tool)
        return _item

    def filter_pair(self, key_phrases: str, ignore: List[str] = None) -> List[Function]:
        """
        过滤group中的函数
        """
        if ignore is None:
            ignore = []
        function_list = []
        for func_name, pair_cls in self.pair_function.items():
            if pair_cls.tool().func_message(message_text=key_phrases):
                if func_name in ignore:
                    continue
                function_list.append(pair_cls.function)
        return function_list



# -*- coding: utf-8 -*-
# @Time    : 2023/7/10 下午9:40
# @Author  : sudoskys
# @File    : event.py
# @Software: PyCharm

def help_message():
    return """
    /help - 帮助
    /chat - 聊天
    /task - 任务
    /tool - 工具列表
    /clear - 删除自己的记录
    
Private Chat Only:
    /bind - 绑定可选平台
    /unbind - 解绑可选平台
    /rset_endpoint - 自定义后端
    /rset_key - 设置openai
    /clear_rset - 抹除自定义设置
    """



# -*- coding: utf-8 -*-

import cjieba


def cut_words_weights(content):
    """
    根据jieba分词，提取关键词及其权重
    :param content:
    :return:
    """
    # jieba提取关键词及其权重
    # 设置停用词
    # jieba.analyse.set_stop_words('path_of_stopwords')
    # tags = jieba.analyse.extract_tags(content, topK=20, withWeight=True)
    tags = cjieba.extract(text=content, top_k=20, with_weight=True)
    tags = [(keyword, int(weight * 10)) for keyword, weight in tags]
    return tags


def hash_keyword_add_weight(keyword_weight, len_hash=64):
    """
    对关键词进行hash, 然后加权
    :param keyword_weight:
    :param len_hash:
    :return:
    """
    # 关键词hash
    keyword_weight = [(bin(hash(keyword)).replace("0b", "").replace("-", "").zfill(len_hash)[-1 * len_hash:], weight)
                      for keyword, weight in keyword_weight]
    # 加权
    add_weight = [0] * len_hash
    for keyword, weight in keyword_weight:
        for i in range(len_hash):
            if keyword[i] == "1":
                add_weight[i] += weight
            else:
                add_weight[i] += -1 * weight
    result = ""
    for _ in add_weight:
        if _ >= 0:
            result += "1"
        else:
            result += "0"
    return result


def cal_hamming_distance(hash_file1, hash_file2):
    """
    计算两篇文章的海明距离
    :param hash_file1:
    :param hash_file2:
    :return:
    """
    hamming_dis = 0
    for i in range(len(hash_file1)):
        if hash_file1[i] != hash_file2[i]:
            hamming_dis += 1
    # print("海明距离：", hamming_dis)
    return hamming_dis


class SimHashSimilarity(object):
    def __init__(self):
        pass

    def run_simhash(self, str1, str2):
        """
        主程序
        :param str1:
        :param str2:
        :return:
        """
        tags1 = cut_words_weights(str1)
        tags2 = cut_words_weights(str2)
        hash_file1 = hash_keyword_add_weight(tags1)
        hash_file2 = hash_keyword_add_weight(tags2)
        hamming_dis = cal_hamming_distance(hash_file1, hash_file2)
        return hamming_dis


if __name__ == "__main__":
    simhash = SimHashSimilarity()
    print(simhash.run_simhash("你妈妈叫你回家吃饭了，小明", "小明的妈妈让你回家吃饭了"))



# -*- coding: utf-8 -*-
# @Time    : 2023/8/21 下午11:27
# @Author  : sudoskys
# @File    : utils.py
# @Software: PyCharm
from urllib.parse import urlparse


def parse_command(command):
    if not command:
        return None, None
    parts = command.split(" ", 1)
    if len(parts) > 1:
        return parts[0], parts[1]
    elif len(parts) == 1:
        return parts[0], None
    else:
        return None, None


def is_valid_url(url):
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False
